#    Copyright (C) 2021 Andrei Puchko
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from decimal import Decimal
import datetime
import hashlib
import secrets
import binascii
import re


def is_sub_list(sublst, lst):
    return len([x for x in sublst if x in lst]) == len(sublst)


def int_(toInt):
    try:
        return int(f"{toInt}")
    except Exception:
        return int(num(toInt))


def float_(toFloat):
    try:
        return float(f"{toFloat}")
    except Exception:
        return float(num(toFloat))


def num(tonum):
    try:
        return Decimal(f"{tonum}")
    except Exception:
        return 0


def nums(number_str):
    number_str = f"{number_str}"
    decimal_sep_pos = max(number_str.rfind(","), number_str.rfind("."))

    if decimal_sep_pos == -1:
        return number_str

    integer_part = number_str[:decimal_sep_pos]
    decimal_part = number_str[decimal_sep_pos + 1:]

    integer_part = re.sub(r"[^0-9-]", "", integer_part)
    integer_part = re.sub(r"(?<!^)-", "", integer_part)

    formatted_number = f"{integer_part}.{decimal_part}"
    return formatted_number


def today():
    return f"{datetime.date.today()}"


class dotdict(dict):
    """dot.notation access to dictionary attributes"""

    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__


def set_dict_default(_dict, _key, _value):
    """set only key does not exist"""

    if not _dict.get(_key):
        _dict[_key] = _value


class Q2Crypto:
    def __init__(self, pin, lenght=256):
        self._pin = pin
        self.pin = [x for x in hashlib.shake_256(self._pin.encode()).hexdigest(lenght // 2)]

    def encrypt(self, text):
        if not isinstance(text, str):
            return None
        text = binascii.hexlify((text + chr(1) + self._pin).encode())
        step = max(len(self.pin) // max(len(text), 1), 1)
        crypted = self.pin[:]
        r = 0
        for x in range(len(text)):
            r += secrets.randbelow(step) + 1
            if r >= len(crypted):
                raise BaseException("Cryptgraphy Error - given text to long")
            while True:
                if chr(text[x]) != crypted[r]:
                    break
                else:
                    r += 1
            crypted[r] = chr(text[x])
        return "".join(crypted)

    def _decrypt(self, crypted):
        if len(crypted) != len(self.pin):
            return None
        rez = []
        for x in range(len(self.pin)):
            if self.pin[x] != crypted[x]:
                rez.append(crypted[x])
        rez = ("".join(rez)).encode()
        try:
            return binascii.unhexlify(rez).decode()
        except Exception as errow:
            return None

    def decrypt(self, crypted):
        if not isinstance(crypted, str):
            return None
        rez = self._decrypt(crypted)
        if rez is not None:
            if rez.endswith(chr(1) + self._pin):
                rez = rez[: -(len(self._pin) + 1)]
            else:
                return None
        return rez

    def check_pin(self, crypted):
        if not isinstance(crypted, str):
            return None
        rez = self._decrypt(crypted)
        if rez is not None:
            if rez.endswith(chr(1) + self._pin):
                return True
            else:
                return None
        return rez



#    Copyright © 2021 Andrei Puchko
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import sys


import q2gui.q2app as q2app
from configparser import ConfigParser

# from q2gui.q2window import Q2Window
from q2gui.q2utils import num

import re
import io
import time
import os
import codecs
import logging


q2_app = None
engine = ""
Q2Form = None

ASK_REMOVE_RECORD_TEXT = "You are about to remove current record! Are You Sure?"
ASK_REMOVE_RECORDS_TEXT = "You are about to remove records<b>(%s)</b>! Are You Sure?"
REMOVE_RECORD_ERROR_TEXT = "Remove record error"

DATE_FORMAT_STRING = "%d.%m.%Y"


ACTION_HIDDEN_ROW_TEXT = "Rows"
ACTION_HIDDEN_ROW_ICON = "eye-off.svg"

ACTION_HIDE_ROW_TEXT = "Hide row"
ACTION_HIDE_ROW_ICON = "list-minus.svg"

ACTION_SHOW_ROW_TEXT = "Show row"
ACTION_SHOW_ROW_ICON = "list-plus.svg"

ACTION_SHOW_NOTHIDDEN_TEXT = "Show not hidden"
ACTION_SHOW_NOTHIDDEN_ICON = "list-chevrons-down-up.svg"

ACTION_SHOW_ALL_TEXT = "Show all"
ACTION_SHOW_ALL_ICON = "list-chevrons-up-down.svg"

ACTION_SHOW_HIDDEN_TEXT = "Show hidden"
ACTION_SHOW_HIDDEN_ICON = "eye.svg"

ACTION_VIEW_TEXT = "View"
ACTION_VIEW_ICON = "row-view.png"
ACTION_VIEW_HOTKEY = "F12"

ACTION_NEW_TEXT = "New"
ACTION_NEW_ICON = "row-new.png"
ACTION_NEW_HOTKEY = "Ins"

ACTION_COPY_TEXT = "Copy"
ACTION_COPY_ICON = "row-copy.png"
ACTION_COPY_HOTKEY = "Ctrl+Ins"

ACTION_EDIT_TEXT = "Edit"
ACTION_EDIT_ICON = "row-edit.png"
ACTION_EDIT_HOTKEY = "Spacebar"

ACTION_REMOVE_TEXT = "Remove"
ACTION_REMOVE_ICON = "row-remove.png"
ACTION_REMOVE_HOTKEY = "Delete"


ACTION_FIRST_ROW_TEXT = "First"
ACTION_FIRST_ROW_ICON = "go-top.png"
ACTION_FIRST_ROW_HOTKEY = "Ctrl+Up"

ACTION_PREVIOUS_ROW_TEXT = "Previous"
ACTION_PREVIOUS_ROW_ICON = "go-up.png"

ACTION_REFRESH_TEXT = "Refresh"
ACTION_REFRESH_ICON = "refresh.png"
ACTION_REFRESH_HOTKEY = "F5"

ACTION_RENUMBER_TEXT = "Renumber"
ACTION_RENUMBER_ICON = ""
ACTION_RENUMBER_HOTKEY = ""

ACTION_NEXT_ROW_TEXT = "Next"
ACTION_NEXT_ROW_ICON = "go-down.png"

ACTION_LAST_ROW_TEXT = "Last"
ACTION_LAST_ROW_ICON = "go-bottom.png"
ACTION_LAST_ROW_HOTKEY = "Ctrl+Down"

ACTION_TOOLS_TEXT = "Tools"
ACTION_TOOLS_ICON = "tools.png"

ACTION_TOOLS_COLOR_TEXT = "Set colors"
ACTION_TOOLS_COLOR_ICON = "color"

ACTION_TOOLS_EXPORT_TEXT = "Export"
ACTION_TOOLS_EXPORT_ICON = "export.png"

ACTION_TOOLS_IMPORT_TEXT = "Import"
ACTION_TOOLS_IMPORT_ICON = "import.png"

ACTION_TOOLS_IMPORT_CLIPBOARD_TEXT = "Paste clipboard"
ACTION_TOOLS_IMPORT_CLIPBOARD_ICON = "paste-csv.png"

ACTION_TOOLS_BULK_UPDATE_TEXT = "Bulk update"
ACTION_TOOLS_BULK_UPDATE_ICON = "bulk-update.png"

ACTION_TOOLS_INFO_TEXT = "Info"
ACTION_TOOLS_INFO_ICON = "info.png"

ACTION_SELECT_TEXT = "Select"
ACTION_SELECT_ICON = "select.png"

ACTION_CLOSE_TEXT = "Close"
ACTION_CLOSE_ICON = "exit.png"

CRUD_BUTTON_EDIT_TEXT = "Edit"
CRUD_BUTTON_EDIT_MESSAGE = "enable editing"

CRUD_BUTTON_OK_TEXT = "OK"
CRUD_BUTTON_OK_MESSAGE = "save data"

CRUD_BUTTON_CANCEL_TEXT = "Cancel"
CRUD_BUTTON_CANCEL_MESSAGE = "Do not save changes"

GRID_ACTION_TEXT = "☰"
GRID_ACTION_ICON = "menu.png"

ARROW_UP_ICON = "arrow-up.png"
ARROW_DOWN_ICON = "arrow-down.png"

FINANCIAL_FORMAT = r"{:,.%sf}"
GRID_COLUMN_WIDTH = 25

MESSAGE_SORTING = "Sorting..."
MESSAGE_SORTING = "Sorting..."
MESSAGE_ROWS_HIDING = "Toggle hide/show rows"
MESSAGE_ROWS_COLOR = "Colorizing rows"

MESSAGE_GRID_DATA_EXPORT_TITLE = "Export data"
MESSAGE_GRID_DATA_EXPORT_WAIT = "Export data to: %s"
MESSAGE_GRID_DATA_EXPORT_ERROR = "Export error: %s"
MESSAGE_GRID_DATA_EXPORT_DONE = "Export done:<br>Rows: %s<br>Time: %.2f sec."

MESSAGE_GRID_DATA_IMPORT_TITLE = "Import data"
MESSAGE_GRID_DATA_IMPORT_WAIT = "Import data to: %s"
MESSAGE_GRID_DATA_IMPORT_ERROR = "Import error: %s"
MESSAGE_GRID_DATA_IMPORT_DONE = "Import done:<br>Rows: %s<br>Time: %.2f sec."

GRID_DATA_INFO_TABLE = "Table/Query"
GRID_DATA_INFO_SQL = "Query"
GRID_DATA_INFO_ROWS = "Rows"
GRID_DATA_INFO_ORDER = "Order"
GRID_DATA_INFO_FILTER = "Filter"
GRID_DATA_INFO_COLUMNS = "Columns"

PASTE_CLIPBOARD_WAIT = "Paste rows"
PASTE_CLIPBOARD_TITLE = "Paste (Clipboard)"
PASTE_CLIPBOARD_FIRST_ROW = "First row is a header"
PASTE_CLIPBOARD_CLIPBOARD_DATA = "Clipboard data"
PASTE_CLIPBOARD_TARGET = "Target"
PASTE_CLIPBOARD_SOURCE = "Source"
PASTE_CLIPBOARD_TARGET_COLUMNS = "Target columns"
PASTE_CLIPBOARD_SOURCE_COLUMNS = "Source columns"

BULK_DATA_ENTRY_TITLE = "Bulk data"
BULK_DATA_MAIN_TITLE = "Bulk update selected rows"
BULK_DATA_WAIT = "Bulk update rows"
BULK_TARGET_TITLE = "Target"
BULK_TARGET_COLUMNS = "Columns"
BULK_TARGET_SELECTED = "Selected"

ASK_ROWS_REMOVING_ERRORS_SUPRESS = "Do not show next errors"
ASK_COPY_CHILD_DATA = "Copy %s?"

DIALOG_OPEN_FILE_TITLE = "Open file"
DIALOG_SAVE_FILE_TITLE = "Save file"

from q2gui.q2style import Q2Style


def load_q2engine(glo, engine="PyQt6"):
    from q2gui.pyqt6.q2app import Q2App as Q2App
    from q2gui.pyqt6.q2form import Q2Form as Q2Form
    from q2gui.pyqt6.q2style import Q2Style as Q2Style

    q2app.engine = engine
    Q2App.Q2Form = Q2Form

    glo["Q2App"] = Q2App
    glo["Q2Form"] = Q2Form
    glo["Q2Style"] = Q2Style


class Q2Heap:
    pass


class Q2Actions(list):
    def __init__(self, action=None):
        self.show_main_button = True
        self.show_actions = True
        self.main_button_text = GRID_ACTION_TEXT
        if isinstance(action, list):
            # self.action_list = action[:]
            self.extend(action[:])
        # else:
        #     self.action_list = []

    def run(self, text):
        for action in self:
            if text == action["text"]:
                action["_worker"]()

    def set_visible(self, text, mode=True):
        for action in self:
            if text == action["text"]:
                action["_set_visible"](mode)
            # elif text == action.get("parent_action_text"):
            #     action["_set_visible_parent_action"](mode)

    def set_disabled(self, text="", mode=True):
        for action in self:
            if text == action["text"]:
                action["_set_disabled"](mode)

    def set_enabled(self, text="", mode=True):
        for action in self:
            if text == action["text"]:
                action["_set_enabled"](mode)

    def add_action(
        self,
        text,
        worker=None,
        icon="",
        mess="",
        hotkey="",
        tag="",
        eof_disabled="",
        child_form=None,
        child_where="",
        child_copy_mode=1,
        child_noshow=0,
    ):
        """ "/view", "/crud" """
        for x in range(len(self)):
            # if text in self[x]["text"] and text.strip()[-1] != "-":
            if text == self[x]["text"] and text.strip()[-1] != "-":
                self[x]["worker"] = worker
                self[x]["hotkey"] = hotkey
                return True
        action = {}
        action["text"] = text
        action["worker"] = worker
        action["child_copy_mode"] = child_copy_mode
        action["child_noshow"] = child_noshow

        if tag == "select":
            icon = ACTION_SELECT_ICON

        # icon = q2_app.get_icon(icon)

        # action["icon"] = icon if os.path.isfile(icon) else ""
        action["icon"] = icon if icon else ""
        action["mess"] = mess
        action["hotkey"] = hotkey
        action["tag"] = tag
        action["eof_disabled"] = eof_disabled
        action["child_form"] = child_form
        action["child_where"] = child_where
        self.append(action)
        return True

    # def insertAction(
    #     self, before, text, worker=None, icon="", mess="", key="", **kvargs
    # ):
    #     for x in self.addAction.__code__.co_varnames:
    #         if x not in ["kvargs", "self"]:
    #             kvargs[x] = locals()[x]
    #     self.action_list.insert(before, kvargs)

    # def removeAction(self, text):
    #     actionIndex = safe_index([x["text"] for x in self.action_list], text)
    #     if actionIndex is not None:
    #         self.action_list.pop(actionIndex)
    pass


class Q2Controls(list):
    class _C:
        def __init__(self, controls):
            self.controls = controls

        def __getattr__(self, name):
            for line in self.controls:
                if line.get("column") == name or line.get("tag") == name:
                    return line
            return None
            # return [line["column"] for line in self.controls]

    def __init__(self):
        self.c = self._C(self)

    def get(self, name):
        return self.c.__getattr__(name)

    def delete(self, name):
        c = self.get(name)
        self.pop(self.index(c))

    def get_names(self):
        return [line["column"] for line in self]

    # def __getitem__(self, list_index):
    #     if isinstance(list_index, str):  # not index but name - return index for name
    #         for x in range(len(self)):
    #             if list_index == self[x].get("column"):
    #                 return x
    #         return None
    #     else:
    #         return super().__getitem__(list_index)

    def add_control(
        self,
        column="",
        label="",
        gridlabel="",
        control="",
        pic="",
        data="",
        datatype="char",
        datalen=0,
        datadec=0,
        pk="",
        ai="",
        migrate="*",
        actions=[],
        alignment=-1,
        to_table="",
        to_column="",
        to_form=None,
        related="",
        db=None,
        mask="",
        opts="",
        when=None,
        show=None,
        valid=None,
        changed=None,
        dblclick=None,
        readonly=None,
        disabled=None,
        check=None,
        noform=None,
        nogrid=None,
        widget=None,
        margins=None,
        stretch=0,
        mess="",
        tag="",
        eat_enter=None,
        hotkey="",
        style="",
        **args,
    ):
        meta = locals().copy()
        del meta["self"]
        meta["_control"] = None
        # meta = self.validate(meta)
        self.append(meta)
        return True

    @staticmethod
    def validate(meta):
        if meta.get("margins") is None:
            if q2app.q2_app:
                meta["margins"] = [
                    q2app.q2_app.content_margin_top,
                    q2app.q2_app.content_margin_right,
                    q2app.q2_app.content_margin_bottom,
                    q2app.q2_app.content_margin_left,
                ]
            else:
                meta["margins"] = [1, 1, 1, 1]

        if meta.get("datatype") == "char":
            if re.match(".*code.*|.*text.*", meta.get("control", ""), re.RegexFlag.IGNORECASE):
                meta["datatype"] = "text"

        if meta.get("datatype", "").lower() == "date":
            meta["control"] = "date"
            meta["datalen"] = 16

        if meta.get("datatype", "").lower() == "time":
            meta["control"] = "time"
            meta["datalen"] = 10

        if meta.get("column").startswith("/"):
            meta["control"] = ""
        elif not meta.get("control") and not meta.get("widget") and meta.get("column"):
            meta["control"] = "line"
            # meta["control"] = ""

        if num(meta.get("datalen", 0)) == 0 and meta.get("control", "") in ("line", "radio"):
            if meta.get("datatype", "").lower()[:3] == "int":
                meta["datalen"] = 9
            elif meta.get("datatype", "").lower() == "bigint":
                meta["datalen"] = 17
            else:
                meta["datalen"] = 100

        if (
            re.match(".*text.*", meta.get("datatype", ""), re.RegexFlag.IGNORECASE)
            and "code" not in meta["control"]
            and "image" not in meta["control"]
        ):
            meta["datalen"] = 0
            meta["control"] = "text"

        if "***" == "".join(["*" if meta.get(x) else "" for x in ("to_table", "to_column", "related")]):
            meta["relation"] = True

        if re.match(".*int.*|.*dec.*|.*num.*", meta.get("datatype", ""), re.RegexFlag.IGNORECASE):
            meta["num"] = True
            if meta.get("pic", "") == "":
                meta["pic"] = "9" * int(num(meta["datalen"]) - num(meta["datadec"])) + (
                    "" if num(meta["datadec"]) == 0 else "." + "9" * int(num(meta["datadec"]))
                )
            if num(meta.get("alignment", -1)) in (
                -1,
                0,
            ):
                meta["alignment"] = 9

        if not meta["column"].startswith("/"):
            if "char" in meta.get("datatype", "") and num(meta.get("datalen")) == 0:
                if meta.get("control") in ("check"):
                    meta["datalen"] = 1
                elif meta.get("control") in ("line"):
                    meta["datalen"] = 100
        return meta


class Q2Settings:
    def __init__(self, filename=""):
        self.filename = filename if filename else "q2gui.ini"
        self.config = ConfigParser()
        self.read()

    def read(self):
        if self.filename in ("none", "memory"):
            self.filename = io.StringIO("")
        if isinstance(self.filename, io.StringIO):
            self.config.read_file(self.filename)
        else:
            if not os.path.isfile(self.filename):
                self.write()
            try:
                self.config.read_file(codecs.open(self.filename, "r", "utf8"))
            except Exception as read_error:
                logging.error(f"Error reading {self.filename}:\n{read_error}")
                self.fix_ini_headers()
                self.config.read_file(codecs.open(self.filename, "r", "utf8"))
            except Exception as recovery_error:
                logging.error(f"Error while trying recover {self.filename}:\n{recovery_error}")
                raise

    def write(self):
        if self.filename == "none":
            return
        if isinstance(self.filename, io.StringIO):
            self.config.write(self.filename)
        else:
            with codecs.open(self.filename, "w", "utf8") as configfile:
                self.config.write(configfile)

    def prepSection(self, section):
        return re.sub(r"\[.*\]", "", section).strip().split("\n")[0].replace("\n", "").strip()

    def get(self, section="", key="", defaultValue=""):
        section = self.prepSection(section)
        key = self.prepSection(key)
        try:
            return self.config.get(section, key)
        except Exception:
            return defaultValue

    def set(self, section="", key="", value=""):
        if section == "":
            return
        section = self.prepSection(section)
        key = self.prepSection(key)
        value = "%(value)s" % locals()
        if not self.config.has_section(section):
            self.config.add_section(section)
        self.config.set(section, key, value)

    def fix_ini_headers(self):
        with open(self.filename, "r", encoding="utf-8") as file:
            content = file.read()

        # fix section headers with sections with hyphens
        pattern = r"\[(.+?)\n\.(.*?)\]"
        fixed_content = re.sub(pattern, r"[\1 \2]", content)

        with open(self.filename, "w", encoding="utf-8") as file:
            file.write(fixed_content)


class Q2Stdout:
    def __init__(self, q2_app):
        self.q2_app = q2_app

    def write(self, output):
        if self.q2_app.dev_mode and hasattr(self.q2_app.stdout_widget, "write"):
            self.q2_app.stdout_widget.write(output)
        else:
            try:
                sys.__stdout__.write(output)
            except Exception as e:
                logging.warning(f"{e}")

    def flush(self):
        sys.__stdout__.flush()


class Q2App:
    Q2Style = Q2Style
    Q2Form = None

    def __init__(self, title=""):
        q2app.q2_app = self
        self.window_title = title
        self.heap = Q2Heap()
        self.db = None
        self.style_file = ""
        self.settings_file = ""
        self.dev_mode = None
        self.settings_file = self.get_argv("ini")
        self.icon = None
        sys.stdout = Q2Stdout(self)

        self.menu_list = []
        self._main_menu = {}

        self.settings = Q2Settings(self.settings_file)

        self.style_file = self.get_argv("style")
        if self.style_file == "":
            self.style_file = "q2gui.qss"

        self.q2style: Q2Style = self.Q2Style(self)
        # print(self.q2style)
        # self.q2style._font_size = 25
        # self.q2style._font_name = "Arial"
        # self.q2style.set_style_sheet(self, "light")
        # self.set_color_mode("light")

        self.menu_list = []
        self.content_margin_top = 0
        self.content_margin_right = None
        self.content_margin_bottom = None
        self.content_margin_left = None
        self.assets_folder = "assets"
        self.set_icon(f"{self.assets_folder}/q2gui.ico")

        self.on_init()

    def set_clipboard(self, text):
        pass

    def get_stdout_height(self):
        pass

    def subwindow_count_changed(self):
        pass

    def disable_menu(self, menu_path=""):
        pass

    def enable_menu(self, menu_path=""):
        pass

    def set_color_mode(self, color_mode=None):
        self.q2style.set_color_mode(self, color_mode)

    def get_color_mode(self):
        return self.q2style.get_color_mode()

    def set_font(self, font_name, font_size):
        pass

    def set_style_sheet(self):
        pass

    def get_icon(self, icon):
        icon_path = f"{self.assets_folder}/{icon}"
        if os.path.isfile(icon_path):
            return icon_path

    def get_argv(self, argtext: str):
        for x in sys.argv:
            if x.startswith(f"/{argtext}:") or x.startswith(f"-{argtext}:"):
                file_name = x[(len(argtext) + 2) :]  # noqa: E203
                return file_name
        return ""

    def add_menu(self, text="", worker=None, before=None, toolbar=None, icon=None, tag=""):
        if text.endswith("|"):
            text = text[:-1]
        if text.startswith("|"):
            text = text[1:]
        self.menu_list.append(
            {"TEXT": text, "WORKER": worker, "BEFORE": before, "TOOLBAR": toolbar, "ICON": icon, "TAG": tag}
        )

    def clear_menu(self):
        self.menu_list = []

    def get_autocompletition_list(self):
        return [
            "test_autocompletion",
            "test_table",
            "test_table.column1",
            "test_table.column2",
        ]

    def build_menu(self):
        self.menu_list = self.reorder_menu(self.menu_list)

    def reorder_menu(self, menu):
        tmp_list = [x["TEXT"] for x in menu]
        tmp_dict = {x["TEXT"]: x for x in menu}
        re_ordered_list = []
        for x in tmp_list:
            # add node element for menu
            menu_node = "|".join(x.split("|")[:-1])
            if menu_node not in re_ordered_list:
                re_ordered_list.append(menu_node)
                tmp_dict[menu_node] = {
                    "TEXT": menu_node,
                    "WORKER": None,
                    "BEFORE": None,
                    "TOOLBAR": None,
                }
            if tmp_dict[x].get("BEFORE") in re_ordered_list:
                re_ordered_list.insert(re_ordered_list.index(tmp_dict[x].get("BEFORE")), x)
            else:
                re_ordered_list.append(x)
        return [tmp_dict[x] for x in re_ordered_list]

    def close(self):
        self.save_geometry(self.settings)

    def save_geometry(self, settings):
        pass

    def restore_geometry(self, settings):
        pass

    def show_statusbar_mess(self, text=""):
        pass

    def clear_statusbar(self):
        pass

    def get_statusbar_mess(self):
        pass

    def show_form(self, form=None, modal="modal"):
        pass

    def get_dpi(self):
        pass

    def focus_changed(self, from_widget, to_widget):
        if not (hasattr(from_widget, "meta") and hasattr(to_widget, "meta")):
            return
        if from_widget.__class__.__name__ in (
            "q2line",
            "q2text",
            "q2relation",
            "q2ScriptEdit",
            "q2ScriptSqlEdit",
        ):
            if from_widget.meta.get("form") == to_widget.meta.get("form"):
                if from_widget.valid() is False:
                    from_widget.set_focus()
        if to_widget.__class__.__name__ in (
            "q2line",
            "q2text",
            "q2relation",
            "q2ScriptEdit",
            "q2ScriptSqlEdit",
        ):
            if from_widget.meta.get("form") == to_widget.meta.get("form"):
                if to_widget:
                    to_widget.when()

    def get_clipboard_text(self):
        pass

    def set_clipboard_text(self):
        pass

    def lock(self):
        pass

    def unlock(self):
        pass

    def set_icon(self):
        pass

    def process_events(self):
        pass

    def on_init(self):
        self.set_icon(f"{self.assets_folder}/q2gui.ico")

    def on_start(self):
        pass

    def on_new_tab(self):
        pass

    def show_menubar(self, mode=True):
        pass

    def hide_menubar(self, mode=True):
        if mode:
            self.show_menubar(False)
        else:
            self.show_menubar(True)

    def is_menubar_visible(self):
        pass

    def is_menubar_enabled(self):
        pass

    def show_toolbar(self, mode=True):
        pass

    def hide_toolbar(self, mode=True):
        if mode:
            self.show_toolbar(False)
        else:
            self.show_toolbar(True)

    def is_toolbar_visible(self):
        pass

    def is_toolbar_enabled(self):
        pass

    def disable_toolbar(self, mode=True):
        pass

    def enable_toolbar(self, mode=True):
        self.disable_toolbar(not mode)
        pass

    def disable_menubar(self, mode=True):
        pass

    def enable_menubar(self, mode=True):
        self.disable_menubar(not mode)

    def disable_tabbar(self, mode=True):
        pass

    def enable_tabbar(self, mode=True):
        self.disable_tabbar(not mode)

    def show_tabbar(self, mode=True):
        pass

    def get_current_q2_form(self):
        pass

    def disable_current_form(self, mode=True):
        pass

    def get_tabbar_text(self):
        pass

    def set_tabbar_text(self, text=""):
        pass

    def hide_tabbar(self, mode=True):
        if mode:
            self.show_tabbar(False)
        else:
            self.show_tabbar(True)

    def is_tabbar_visible(self):
        pass

    def is_tabbar_enabled(self):
        pass

    def show_statusbar(self, mode=True):
        pass

    def keyboard_modifiers(self):
        pass

    def hide_statusbar(self, mode=True):
        if mode:
            self.show_statusbar(False)
        else:
            self.show_statusbar(True)

    def is_statusbar_visible(self):
        pass

    def get_char_width(self, char="W"):
        return 9

    def get_char_height(self):
        return 9

    def sleep(self, seconds=0):
        time.sleep(seconds)

    @staticmethod
    def get_open_file_dialoq(header="Open file", path="", filter=""):
        pass

    @staticmethod
    def get_save_file_dialoq(header="Save file", path="", filter="", confirm_overwrite=True):
        pass

    def add_new_tab(self):
        pass

    def run(self):
        self.build_menu()

    def click_submenu(menu, submenu_text):
        pass



from q2gui.version import __version__

from q2gui.pyqt6.q2app import Q2App as Q2App
from q2gui.pyqt6.q2form import Q2Form as Q2Form
from q2gui.pyqt6.q2style import Q2Style as Q2Style

from q2gui.q2model import Q2CursorModel as Q2CursorModel
from q2gui.q2dialogs import q2_mess as q2_mess
from q2gui.q2dialogs import q2_ask_yn as q2_ask_yn



#    Copyright © 2021 Andrei Puchko
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


import q2gui.q2app as q2app
from q2gui.q2model import Q2Model, Q2CursorModel
from q2gui.q2utils import int_, num, nums
import re
import html
import json
import gettext

_ = gettext.gettext


VIEW = "VIEW"
NEW = "NEW"
COPY = "COPY"
EDIT = "EDIT"
NO_DATA_WIDGETS = ("button", "toolbutton", "frame", "label")
NO_LABEL_WIDGETS = ("button", "toolbutton", "frame", "label", "check")


class Q2Form:
    def __init__(self, title=""):
        super().__init__()
        self.title = title.strip().split("\n")[0][:50]
        self.name = title
        self.form_stack = []
        self.style_sheet = ""

        self.hide_title = False
        self.maximized = False
        self.init_size = [0, 0]

        self.heap = q2app.Q2Heap()
        self.actions = q2app.Q2Actions()
        self.grid_navi_actions = []
        self.controls = q2app.Q2Controls()
        self.system_controls = q2app.Q2Controls()
        self.model = None
        self.db = None
        self.q2_app = q2app.q2_app
        if hasattr(self.q2_app, "db"):
            self.db = q2app.q2_app.db
        self._model_record = {}  # contains the data of the currently edited record

        # Shortcuts to elements
        self.s = Q2FormData(self)  # widgets data by name
        self.w = Q2FormWidget(self)  # widgets by name
        self.a = Q2FormAction(self)  # Actions by text
        self.r = Q2ModelData(self)  # Grid data by name
        self.c = self.controls.c  # control dict

        self.prev_form = None
        self.children_forms = []  # forms inside this form
        self.i_am_child = None
        self.max_child_level = 1  # max depth for child forms

        self.ok_button = False
        self.cancel_button = False
        self.ok_pressed = None

        self.show_grid_action_top = True
        self.do_not_save_geometry = False

        # Must be redefined in any subclass
        self._Q2FormWindow_class = Q2FormWindow
        self._q2dialogs = None

        self._in_close_flag = False
        self.last_closed_form = None
        self.last_closed_form_widgets_text = {}

        self.grid_form = None
        self.crud_form = None
        self.crud_mode = ""

        self.no_view_action = False

        self.current_row = -1
        self.current_column = -1
        self.last_current_row = -1
        self.last_current_column = -1

        # Must be called in subclass
        # self.on_init()
        pass

    def on_init(self):
        pass

    def run_action(self, text=""):
        for x in self.actions:
            if text == x["text"]:
                x["_worker"]()

    def disable_action(self, text="", mode=True):
        self.actions.set_disabled(text, mode)

    def enable_action(self, text="", mode=True):
        self.actions.set_enabled(text, mode)

    def run(self, order="", where=""):
        if self.model:
            need_refresh = False
            if order:
                need_refresh = True
                self.model.set_order(order)
            if where:
                need_refresh = True
                self.model.set_where(where)
            if need_refresh:
                self.model.refresh()
            self.show_mdi_modal_grid()
        else:
            self.show_mdi_modal_form()
        return self

    def show_progressbar(self, title="", count=""):
        return self._q2dialogs.Q2WaitShow(title, count)

    def show_(self):
        for x in self.widgets():
            if hasattr(self.widgets()[x], "show_"):
                self.widgets()[x].show_()
        self.form_refresh()
        # for x in self.controls:
        #     if x.get("show") and x.get("column") in self.widgets():
        #         self.widgets()[x.get("column")].set_text(x["show"](mode="form"))

    def set_model(self, model):
        self.model: Q2Model = model
        self.model.q2_form = self
        self.model.build()
        return self.model

    def set_cursor(self, cursor):
        self.set_model(Q2CursorModel(cursor))

    def refresh(self, soft=None):
        row = self.current_row
        col = self.current_column
        if soft:
            self.model.refresh()
            self.refresh_children()
        else:
            self._q2dialogs.q2working(
                # lambda: (self.model.refresh(), self.refresh_children()), _("Refreshing...")
                lambda: (self.model.refresh(),), _("Refreshing...")
            )
        self.q2_app.show_statusbar_mess(self.model.row_count())
        self.set_grid_index(row, col)
        if self.model.refreshed:
            #     self.form_refresh()
            self.model.refreshed = False

    def form_refresh(self):
        pass

    def widget(self):
        if self.form_stack:
            return self.form_stack[-1]

    def widgets(self):
        if self.form_stack:
            return self.form_stack[-1].widgets
        else:
            return {}

    def widgets_list(self):
        return [self.form_stack[-1].widgets[x] for x in self.form_stack[-1].widgets]

    def focus_widget(self):
        return q2app.q2_app.focus_widget()

    def close(self):
        if self.form_stack:
            self.last_closed_form = self.form_stack[-1]
            self.save_closed_form_text()
            self.form_stack[-1].close()
            # self.q2_app.process_events()

    def save_closed_form_text(self):
        self.last_closed_form_widgets_text = {
            x: self.last_closed_form.widgets[x].get_text()
            for x in self.last_closed_form.widgets
            if hasattr(self.last_closed_form.widgets[x], "get_text")
        }

    def _close(self, q2form_window=None):
        if self._in_close_flag:
            return
        self._in_close_flag = True
        if self.form_stack:
            self.last_closed_form = self.form_stack[-1]
            self.form_stack[-1].save_splitters()
            self.save_closed_form_text()
        if q2form_window is not None and q2form_window.is_crud is None:
            self.after_form_closed()
        self._in_close_flag = False

    def show_form(self, title="", modal="modal"):
        self.get_form_widget(title).show_form(modal)

    def show_mdi_form(self, title=""):
        form_widget = self.get_form_widget(title)
        form_widget.show_form(modal="")

    def show_mdi_modal_form(self, title=""):
        form_widget = self.get_form_widget(title)
        form_widget.show_form("modal")

    def show_app_modal_form(self, title=""):
        self.get_form_widget(title).show_form(modal="super")

    def show_grid(self, title="", modal=""):
        self.get_grid_widget(title).show_form(modal)

    def show_mdi_grid(self, title=""):
        self.get_grid_widget(title).show_form(modal="")

    def show_mdi_modal_grid(self, title=""):
        self.get_grid_widget(title).show_form(modal="modal")

    def show_app_modal_grid(self, title=""):
        self.get_grid_widget(title).show_form(modal="super")

    def get_form_widget(self, title=""):
        form_widget = self._Q2FormWindow_class(self, title)
        form_widget.build_form()
        return form_widget

    def get_grid_widget(self, title=""):
        self.grid_form = self._Q2FormWindow_class(self, title)
        self.model.build()
        self.get_grid_crud_actions()
        self.before_grid_build()
        self.grid_form.build_grid()
        return self.grid_form

    def get_widget(self):
        if self.model is not None:
            return self.get_grid_widget()
        else:
            return self.get_form_widget()

    def get_grid_crud_actions(self):
        is_crud = self.a.__getattr__("/crud")

        tmp_actions = q2app.Q2Actions()
        if not self.no_view_action:
            self.add_action_view(tmp_actions)
        if is_crud and not self.model.readonly:
            self.add_action_new(tmp_actions)
            self.add_action_copy(tmp_actions)
            self.add_action_edit(tmp_actions)
            tmp_actions.add_action(text="-")
            self.add_action_delete(tmp_actions)
            tmp_actions.add_action(text="-")

        if "seq" in self.model.columns:
            self.add_action_seq()

        for x in self.actions:
            if x.get("text").startswith("/"):
                continue
            tmp_actions.append(x)
        self.actions = tmp_actions

    def before_form_build(self):
        pass

    def before_grid_build(self):
        pass

    def add_action_view(self, actions=None):
        if actions is None:
            actions = self.actions
        actions.add_action(
            text=q2app.ACTION_VIEW_TEXT,
            worker=lambda: self.show_crud_form(VIEW),
            icon=q2app.ACTION_VIEW_ICON,
            hotkey=q2app.ACTION_VIEW_HOTKEY,
            eof_disabled=1,
            tag="view",
        )

    def add_action_delete(self, actions=None):
        if actions is None:
            actions = self.actions
        actions.add_action(
            text=q2app.ACTION_REMOVE_TEXT,
            worker=self.crud_delete,
            icon=q2app.ACTION_REMOVE_ICON,
            hotkey=q2app.ACTION_REMOVE_HOTKEY,
            eof_disabled=1,
            tag="red",
        )

    def add_action_copy(self, actions=None):
        if actions is None:
            actions = self.actions
        actions.add_action(
            text=q2app.ACTION_COPY_TEXT,
            worker=lambda: self.show_crud_form(COPY),
            icon=q2app.ACTION_COPY_ICON,
            hotkey=q2app.ACTION_COPY_HOTKEY,
            eof_disabled=1,
        )

    def add_action_edit(self, actions=None):
        if actions is None:
            actions = self.actions
        actions.add_action(
            text=q2app.ACTION_EDIT_TEXT,
            worker=lambda: self.show_crud_form(EDIT),
            icon=q2app.ACTION_EDIT_ICON,
            hotkey=q2app.ACTION_EDIT_HOTKEY,
            eof_disabled=1,
            tag="edit",
        )

    def add_action_new(self, actions=None):
        if actions is None:
            actions = self.actions
        actions.add_action(
            text=q2app.ACTION_NEW_TEXT,
            worker=lambda: self.show_crud_form(NEW),
            icon=q2app.ACTION_NEW_ICON,
            hotkey=q2app.ACTION_NEW_HOTKEY,
        )

    def add_action_seq(self, actions=None):
        if actions is None:
            actions = self.actions
        actions.add_action(
            "Move up", self.move_seq_up, icon="arrow-up.png", eof_disabled=1, hotkey="Ctrl+Alt+Up"
        )
        actions.add_action(
            "Move down", self.move_seq_down, icon="arrow-down.png", eof_disabled=1, hotkey="Ctrl+Alt+Down"
        )
        actions.add_action(
            q2app.ACTION_RENUMBER_TEXT,
            self.seq_renumber,
            icon=q2app.ACTION_RENUMBER_ICON,
            hotkey=q2app.ACTION_RENUMBER_HOTKEY,
            eof_disabled=1,
        )
        actions.add_action("-")

    def move_seq_up(self):
        selected_row = sorted(self.get_grid_selected_rows())
        if selected_row[0] == 0:
            return

        for row in selected_row:
            self.set_grid_index(row)
            nr = self.model.get_record(self.current_row - 1)
            cr = self.model.get_record(self.current_row)
            if nr["seq"] == cr["seq"]:
                nr["seq"] = "%s" % (int_(nr["seq"]) - 1)
            else:
                nr["seq"], cr["seq"] = cr["seq"], nr["seq"]
            self.model.update(nr)
            self.model.update(cr)
        self.refresh(True)
        self.set_grid_index(selected_row[0])
        self.set_grid_selected_rows([x - 1 for x in selected_row])

    def move_seq_down(self):
        selected_row = sorted(self.get_grid_selected_rows())
        selected_row.reverse()
        if selected_row[0] == self.model.row_count() - 1:
            return

        for row in selected_row:
            self.set_grid_index(row)
            nr = self.model.get_record(self.current_row + 1)
            cr = self.model.get_record(self.current_row)
            if nr["seq"] == cr["seq"]:
                nr["seq"] = "%s" % (int_(nr["seq"]) + 1)
            else:
                nr["seq"], cr["seq"] = cr["seq"], nr["seq"]
            self.model.update(nr)
            self.model.update(cr)
        self.refresh(True)
        self.set_grid_index(selected_row[0])
        self.set_grid_selected_rows([x + 1 for x in selected_row])

    def seq_renumber(self):
        curr_row = self.current_row
        pk = self.model.get_meta_primary_key()
        wait = self._q2dialogs.Q2WaitShow(self.model.row_count(), _("Renumber sequence"))
        for x in range(self.model.row_count()):
            wait.step(1000)
            self.set_grid_index(x)
            # dic = {pk: self.model.get_record(x)[pk]}
            dic = {pk: self.r.__getattr__(pk)}
            dic["seq"] = x + 1
            self.model.update(dic, refresh=False)
        wait.close()
        self.refresh(True)
        self.set_grid_index(curr_row)

    def is_grid_updateable(self):
        table_name = self.model.get_table_name()
        return table_name != "" and " " not in table_name

    def build_grid_view_auto_form(self):
        # Define layout
        if self.model.records:
            self.add_control("/f", "Frame with form layout")
            # Populate it with the columns from csv
            for x in self.model.records[0]:
                self.add_control(x, x, control="line", datalen=100)
            # Assign data source
            self.model.readonly = True
            self.actions.add_action(text="/view", eof_disabled=1)

            if self.model.filterable:

                def run_filter_data_form():
                    filter_form = self.__class__("Filter Conditions")
                    # Populate form with columns
                    for x in self.controls:
                        filter_form.controls.add_control(
                            column=x["column"],
                            label=x["label"],
                            control=x["control"],
                            check=False if x["column"].startswith("/") else True,
                            datalen=x["datalen"],
                        )

                    def before_form_show():
                        # put previous filter conditions to form
                        for x in self.model.get_where().split(" and "):
                            if "' in " not in x:
                                continue
                            column_name = x.split(" in ")[1].strip()
                            column_value = x.split(" in ")[0].strip()[1:-1]
                            filter_form.w.__getattr__(column_name).set_text(column_value)
                            filter_form.w.__getattr__(column_name).check.set_checked()

                    def valid():
                        # apply new filter to grid
                        filter_list = []
                        for x in filter_form.widgets_list():
                            if x.check and x.check.is_checked():
                                filter_list.append(f"'{x.get_text()}' in {x.meta['column']}")
                        filter_string = " and ".join(filter_list)
                        self.model.set_where(filter_string)

                    filter_form.before_form_show = before_form_show
                    filter_form.valid = lambda: self._q2dialogs.q2working(valid, q2app.MESSAGE_SORTING)
                    filter_form.ok_button = 1
                    filter_form.cancel_button = 1
                    filter_form.add_ok_cancel_buttons()
                    filter_form.show_mdi_modal_form()

                self.actions.add_action("Filter", worker=run_filter_data_form, hotkey="F9", eof_disabled=1)

    def get_table_schema(self):
        rez = []
        if self.model is not None:
            table_name = ""
            table_name = self.model.get_table_name()
            for meta in self.controls:
                meta = q2app.Q2Controls.validate(meta)
                if meta["column"].startswith("/"):
                    continue
                if not meta.get("migrate"):
                    continue
                if meta.get("control") in NO_DATA_WIDGETS:
                    continue
                column = {
                    "table": table_name,
                    "column": meta["column"],
                    "datatype": meta["datatype"],
                    "datalen": meta["datalen"],
                    "datadec": meta["datadec"],
                    "to_table": meta["to_table"],
                    "to_column": meta["to_column"],
                    "related": meta["related"],
                    "pk": meta["pk"],
                    "ai": meta["ai"],
                }
                rez.append(column)
        return rez

    def get_current_record(self):
        if self.model:
            return self.model.get_record(self.current_row)

    def _valid(self):
        if self.valid() is False:
            return
        self.ok_pressed = True
        self.close()

    def add_ok_cancel_buttons(self):
        if not self.ok_button and not self.cancel_button:
            return
        buttons = q2app.Q2Controls()
        buttons.add_control("/")
        buttons.add_control("/h", "-")
        buttons.add_control("/s")
        if self.ok_button:
            buttons.add_control(
                column="_ok_button",
                label=q2app.CRUD_BUTTON_OK_TEXT,
                control="button",
                hotkey="PgDown",
                valid=self._valid,
                tag="_ok_button",
            )
        if self.cancel_button:
            buttons.add_control(
                column="_cancel_button",
                label=q2app.CRUD_BUTTON_CANCEL_TEXT,
                control="button",
                mess=q2app.CRUD_BUTTON_CANCEL_MESSAGE,
                valid=self.close,
                tag="_cancel_button",
            )
        buttons.add_control("/")

        self.system_controls = buttons

    def add_crud_buttons(self, mode):
        buttons = q2app.Q2Controls()
        buttons.add_control("/")
        buttons.add_control("/h", "-", tag="crud_buttons")
        if not self.no_view_action:
            buttons.add_control(
                column="_prev_button",
                label="▲",
                control="button",
                mess="Previous row",
                valid=lambda: self.move_crud_view(8),
                disabled=True if mode is not VIEW else False,
                hotkey="PgUp",
            )
            buttons.add_control(
                column="_next_button",
                label="▼",
                control="button",
                mess="Next row",
                valid=lambda: self.move_crud_view(2),
                disabled=True if mode is not VIEW else False,
                hotkey="PgDown",
            )
            buttons.add_control("/s")

            if self.a.tag("edit"):
                buttons.add_control(
                    column="_edit_button",
                    label=q2app.CRUD_BUTTON_EDIT_TEXT,
                    control="button",
                    mess=q2app.CRUD_BUTTON_EDIT_MESSAGE,
                    valid=self.crud_view_to_edit,
                    disabled=True if mode is not VIEW else False,
                )
                buttons.add_control("/s")
        else:
            buttons.add_control("/s")

        buttons.add_control(
            column="_ok_button",
            label=q2app.CRUD_BUTTON_OK_TEXT,
            control="button",
            mess=q2app.CRUD_BUTTON_OK_MESSAGE,
            disabled=True if mode is VIEW else False,
            hotkey="PgDown",
            valid=self.crud_save,
            tag="_ok_button",
        )

        buttons.add_control(
            column="_cancel_button",
            label=q2app.CRUD_BUTTON_CANCEL_TEXT,
            control="button",
            mess=q2app.CRUD_BUTTON_CANCEL_MESSAGE,
            # valid=self.crud_close,
            valid=self.close,
            tag="_cancel_button",
        )
        buttons.add_control("/")
        self.system_controls = buttons

    def crud_view_to_edit(self):
        self.crud_form.set_title(f"{self.title}.[EDIT]")
        self.w._ok_button.set_enabled(True)
        self.w._prev_button.set_enabled(False)
        self.w._next_button.set_enabled(False)
        self.w._edit_button.set_enabled(False)

    def move_crud_view(self, mode):
        """move current grid record
        up (mode=8) or down (mode=2) - look at numpad to understand why
        and update values in crud_form
        """
        self.move_grid_index(mode)
        self.set_crud_form_data()
        self.before_form_show()

    def crud_delete(self):
        selected_rows = self.get_grid_selected_rows()
        if len(selected_rows) == 1:
            ask_text = q2app.ASK_REMOVE_RECORD_TEXT
        else:
            ask_text = q2app.ASK_REMOVE_RECORDS_TEXT % len(selected_rows)
        waitbar = None
        if selected_rows and self._q2dialogs.q2AskYN(ask_text) == 2:
            show_error_messages = True
            if len(selected_rows) > 10:
                waitbar = self.show_progressbar(q2app.MESSAGE_ROWS_REMOVING, len(selected_rows))
            for row in selected_rows:
                if waitbar:
                    waitbar.step(1000)

                if self.before_delete() is False:
                    continue
                if self.model.delete(row, refresh=False) is not True and show_error_messages:
                    if selected_rows.index(row) == len(selected_rows) - 1:
                        self._q2dialogs.q2Mess(self.model.get_data_error())
                    else:
                        if (
                            self._q2dialogs.q2AskYN(
                                q2app.REMOVE_RECORD_ERROR_TEXT
                                + "<br>"
                                + self.model.get_data_error()
                                + "<br>"
                                + q2app.ASK_ROWS_REMOVING_ERRORS_SUPRESS
                            )
                            == 2
                        ):
                            show_error_messages = False
                self.after_delete()
            if waitbar:
                waitbar.close()
            self.model.refresh()
            if self.model.row_count() < 0:
                self.current_row = -1
                self.current_column = -1
            self.set_grid_index(row)
            self.refresh_children()

    def get_grid_selected_rows(self):
        return self.grid_form.get_grid_selected_rows()

    def set_grid_selected_rows(self, index_list):
        return self.grid_form.set_grid_selected_rows(index_list)

    def before_delete(self):
        pass

    def after_delete(self):
        pass

    def crud_save(self, close_form=True):
        if self.before_crud_save() is False:
            return
        crud_data = self.get_crud_form_data()
        if self.crud_mode in [EDIT, VIEW]:
            rez = self.update_current_row(crud_data)
        else:
            rez = self.model.insert(crud_data, self.current_row, refresh=False)
            if self.crud_mode == COPY:
                self.prepare_copy_children_data()
            self.model.refresh()

        if rez:
            self.set_grid_index(self.model.seek_row(crud_data))
        if rez is False:
            self._q2dialogs.q2Mess(self.model.get_data_error())
        else:
            if self.crud_mode == COPY:
                self.copy_children_data()
            self.after_crud_save()
            if close_form:
                self.close()
                self.crud_form = None

    def prepare_copy_children_data(self):
        self.copy_records = {}
        # for action in self.children_forms:
        for pos, action in enumerate(self.children_forms):
            if int_(action["child_copy_mode"]) == 3:
                continue
            if (
                int_(action["child_copy_mode"]) == 1
                and self._q2dialogs.q2AskYN(q2app.ASK_COPY_CHILD_DATA % action["text"]) != 2
            ):
                continue
            self.copy_records[pos] = []
            for row_number in range(action["child_form_object"].model.row_count()):
                source_record = action["child_form_object"].model.get_record(row_number)
                self.copy_records[pos].append(source_record)

    def copy_children_data(self):
        for pos, action in enumerate(self.children_forms):
            if pos not in self.copy_records:
                continue
            for source_record in self.copy_records[pos]:
                action["child_form_object"].model.insert(source_record)

    def update_current_row(self, crud_data):
        rez = self.model.update(crud_data, self.current_row)
        self.set_grid_index(self.current_row)
        return rez

    def get_crud_form_data(self):
        # put data from form into self._model_record
        for x in self.crud_form.widgets:
            if x.startswith("/"):
                continue
            widget = self.crud_form.widgets[x]
            if widget.meta.get("control") in NO_DATA_WIDGETS:
                continue
            self._model_record[x] = self.s.__getattr__(x)

            # if widget.meta.get("check") and not widget.check.get_text():
            if widget.meta.get("check") and not widget.check.is_checked():
                if widget.meta.get("num"):
                    value = "0"
                else:
                    value = ""
                self._model_record[x] = value

        return self._model_record

    def show_crud_form(self, mode, modal="modal"):
        """mode - VIEW, NEW, COPY, EDIT"""
        self.crud_mode = mode
        self.add_crud_buttons(mode)
        self.crud_form = self._Q2FormWindow_class(self, f"{self.title}.[{mode}]")
        self.crud_form.is_crud = True
        self.crud_form.build_form()
        self.set_crud_form_data(mode)
        self.crud_form.show_form(modal=modal)

    def set_crud_form_data(self, mode=EDIT):
        """set current record's value in crud_form"""
        where_string = self.model.get_where()
        where_dict = {}
        if "=" in where_string:
            for part in self.model.get_where().split(" and "):
                eq = part.split("=")
                if len(eq) == 2:
                    if self.controls.get(eq[0].strip()):
                        where_dict[eq[0].strip()] = eq[1].strip()
                else:
                    continue
        else:
            where_dict = {}

        if self.current_row >= 0:
            self.model.refresh_record(self.current_row)
            self._model_record = dict(self.model.get_record(self.current_row))
            for x in self._model_record:
                if x not in self.crud_form.widgets:
                    if mode == NEW:
                        self._model_record[x] = ""
                    continue
                if mode in (NEW, COPY) and x == "seq":
                    self.crud_form.widgets[x].set_text(self.model.get_next_sequence(x, self._model_record[x]))
                if (
                    self.controls.c.__getattr__(x)["pk"]
                    and mode in (NEW, COPY)
                    and not self.controls.c.__getattr__(x)["ai"]
                ):
                    # for new record - get next primary key
                    self.crud_form.widgets[x].set_text(self.model.get_uniq_value(x, self._model_record[x]))

                if self.c.__getattr__(x) is None:
                    if mode == NEW:
                        self._model_record[x] = ""
                    continue
                if self.c.__getattr__(x)["check"]:
                    if self.c.__getattr__(x)["num"]:
                        value = num(self._model_record[x])
                    else:
                        value = self._model_record[x]
                    if value:
                        self.crud_form.widgets[x].check.set_checked(True)
                        # self.crud_form.widgets[x].check.set_text("*")
                if mode == NEW:
                    if x not in where_dict and x != "seq" and not self.controls.c.__getattr__(x)["pk"]:
                        self.crud_form.widgets[x].set_text("")
                        self._model_record[x] = ""
                else:
                    self.crud_form.widgets[x].set_text(self._model_record[x])
        # take care about PK and filters
        for x in self.controls.get_names():
            if x not in self.crud_form.widgets:
                continue
            if mode == EDIT and self.controls.get(x)["pk"] and x in self.crud_form.widgets:
                # Disable primary key when edit
                self.crud_form.widgets[x].set_disabled()
            elif mode == NEW and x in where_dict:
                # set where fields
                if where_dict[x][0] == where_dict[x][-1] and where_dict[x][0] in (
                    '"',
                    "'",
                ):
                    where_dict[x] = where_dict[x][1:-1]  # cut quotes
                self.crud_form.widgets[x].set_text(where_dict[x])
                self.crud_form.widgets[x].set_disabled()

    def _grid_index_changed(self, row, column):
        refresh_children_forms = row != self.current_row and row >= 0
        refresh_children_forms = True
        self.last_current_row = self.current_row
        self.last_current_column = self.current_column
        self.current_row = row
        self.current_column = column
        if refresh_children_forms:
            self.refresh_children()
            self.grid_index_changed()

    def grid_index_changed(self):
        pass

    def refresh_children(self):
        for x in self.actions + self.grid_navi_actions:
            if x.get("engineAction") and "_set_disabled" in x:
                x["_set_disabled"](True if x.get("eof_disabled") and self.model.row_count() <= 0 else False)

        for action in self.children_forms:
            filter = self.get_where_for_child(action)
            action["child_form_object"].model.set_where(filter)
            action["child_form_object"].model.refresh()
            action["child_form_object"].set_grid_index()
            action["child_form_object"].refresh_children()

    def show_child_form(self, action):
        child_form = action.get("child_form")()
        child_form.prev_form = self
        child_form.model.set_where(self.get_where_for_child(action))
        child_form.model.refresh()
        child_form.show_mdi_modal_grid()
        # self.refresh()

    def get_where_for_child(self, action):
        if self.current_row >= 0 and self.model.row_count() > 0:
            current_record = self.model.get_record(self.current_row)

            if action.get("child_form_object"):
                if action.get("child_form_object").grid_form:
                    action["child_form_object"].grid_form.set_enabled()
            return action["child_where"].format(**current_record)
        else:
            if action["child_form_object"].grid_form:
                action["child_form_object"].grid_form.set_disabled()
            return "1=2"

    def grid_header_clicked(self, column, direction=None):
        current_record = self.model.get_record(self.current_row)
        if self.model is not None:
            self._q2dialogs.q2working(lambda: self.model.set_order(column, direction), q2app.MESSAGE_SORTING)
            self.refresh()
            self.set_grid_index(self.model.seek_row(current_record))

    def grid_double_clicked(self):
        for tag in ("select", "view", "edit"):
            action = self.a.tag(tag)
            if action and action.get("worker"):
                action.get("worker")()
                break

    def set_grid_index(self, row=None, column=None):
        if row is None:
            row = self.current_row
        if column is None:
            column = self.current_column
        if self.grid_form:
            self.grid_form.set_grid_index(row, column)

    def move_grid_index(self, mode):
        self.grid_form.move_grid_index(mode)

    def get_controls(self):
        self.add_ok_cancel_buttons()
        self.before_form_build()
        return self.controls + self.system_controls

    def when(self):
        pass

    def valid(self):
        pass

    def before_grid_show(self):
        pass

    def after_grid_show(self):
        pass

    def before_form_show(self):
        pass

    def after_form_show(self):
        pass

    def after_form_closed(self):
        pass

    def before_crud_save(self):
        pass

    def after_crud_save(self):
        pass

    def add_control(
        self,
        column="",
        label="",
        gridlabel="",
        control="",
        pic="",
        data="",
        datatype="char",
        datalen=0,
        datadec=0,
        pk="",
        ai="",
        migrate="*",
        actions=[],
        alignment=-1,
        to_table="",
        to_column="",
        to_form=None,
        related="",
        db=None,
        mask="",
        opts="",
        when=None,
        show=None,
        valid=None,
        changed=None,
        dblclick=None,
        readonly=None,
        disabled=None,
        check=None,
        noform=None,
        nogrid=None,
        widget=None,
        margins=None,
        stretch=0,
        mess="",
        tag="",
        eat_enter=None,
        hotkey="",
        style="",
        **args,
    ):
        """Adds a new control(widget) into the form
        arguments:
        column: name of form variable, also migrates to DB as column name
        label: shows in the form as a label and in the grid as column label
        gridlabel: shows in the grid as column label when not empty
        control: control type - "line", "date", "combo", "list",
                                "radio", "check", "text", "code",
                                "button", "toolbutton"
        pic: some extra data, the meaning depends
            pic="F" if control="line" and datatype="num" - thouthands delimiters
            pic="*" if control="line" - password input
            pic="Item1;Item2;.." if control in ["radio", "list", "combo"] - controls items
        data: initial value for control
        datatype: defines default aligment and input mask
                  if migrate=True - defines sql datatype
        datalen=0: defines lenght of control and sql datatype
        datadec=0: defines mask and sql datatype
        pk: if migrate=True - this column is primary key
        ai: if migrate=True - this column is autoincrement
        migrate="*": if migrate=True - this column will be migrated to sql
        actions: actions (Q2Actions) for widget
        alignment: numbers [1-9] - see see keyboards numpad for meanings

        to_table: this column linked to table (Foreign key)
        to_column: this column linked to to_table.to_column (Foreign key)
        to_form: Q2Form object, that appears to search and select the value of to_column
        related: displayed value from table to_table (sql expression)

        db: explicit database(Q2Db)
        mask: -
        opts: -
        when: focus in callback
        show: -
        valid: focus out, data changed callback
        dblclick: grid double click callback
        readonly: .
        disabled: .
        check: has checkbox, contriol is enabled when it is checked
        noform: do not show this control in the form
        nogrid: do not show this control in the grid
        widget: external widget (QWidget, ....)
        margins: .
        stretch: .
        mess: tooltip message
        tag: extradata, sometimes used as name, sometimes as a color
        eat_enter: by default Enter key works like Tab
        hotkey: usually for control="button", for example "Ok" - "PgDown"
        """

        if isinstance(column, dict):
            self.controls.add_control(**column)
        else:
            d = locals().copy()
            del d["self"]
            self.controls.add_control(**d)
        return True  # Do not delete - it allows indentation in code

    def add_action(
        self,
        text,
        worker=None,
        icon="",
        mess="",
        hotkey="",
        tag="",
        eof_disabled="",
        child_form=None,
        child_where="",
        child_copy_mode=1,
        child_noshow=0,
    ):
        """
        child_form - form class or function(fabric) that returns form object
        """
        d = locals().copy()
        del d["self"]
        self.actions.add_action(**d)

    def validate_impexp_file_name(self, file, filetype):
        ft = re.split(r"[^\w]", filetype)[0].lower()
        filetype = f".{ft}"
        file += "" if file.lower().endswith(filetype) else filetype
        return file

    def hidden_row_hide(self):
        self.hidden_row_toggle("*")

    def hidden_row_show(self):
        self.hidden_row_toggle("")

    def hidden_row_toggle(self, status):
        selected_rows = self.get_grid_selected_rows()
        if selected_rows:
            pk = self.model.get_meta_primary_key()
            waitbar = None
            if len(selected_rows) > 10:
                waitbar = self.show_progressbar(q2app.MESSAGE_ROWS_HIDING, len(selected_rows))
            for row in selected_rows:
                if waitbar:
                    waitbar.step(1000)
                dic = {pk: self.model.get_record(row)[pk]}
                dic["q2_hidden"] = status
                self.model.update(dic, self.current_row, refresh=False)
            self.model.refresh()
            self.set_grid_index(self.current_row)
            if waitbar:
                waitbar.close()

    def hidden_row_show_not_hidden(self):
        self.model.set_hidden_row_status("show_not_hidden")
        self.refresh()

    def hidden_row_show_all(self):
        self.model.set_hidden_row_status("show_all")
        self.refresh()

    def hidden_row_show_hidden(self):
        self.model.set_hidden_row_status("show_hidden")
        self.refresh()

    def grid_data_export(self):
        file, filetype = q2app.q2_app.get_save_file_dialoq(
            q2app.MESSAGE_GRID_DATA_EXPORT_TITLE, filter="CSV (*.csv);;JSON(*.json)"
        )
        if not file:
            return
        file = self.validate_impexp_file_name(file, filetype)
        waitbar = self._q2dialogs.Q2WaitShow(
            q2app.MESSAGE_GRID_DATA_EXPORT_WAIT % file, self.model.row_count()
        )
        try:
            self.model.data_export(file, tick_callback=lambda: waitbar.step(1000))
        except Exception as e:
            self._q2dialogs.q2Mess(q2app.MESSAGE_GRID_DATA_EXPORT_ERROR % (file + ": " + str(e)))
            waitbar.close()
        else:
            _count, _time = waitbar.close()
            self._q2dialogs.q2Mess(q2app.MESSAGE_GRID_DATA_EXPORT_DONE % (_count, _time))
        waitbar.close()

    def grid_data(self, row, col, skip_format=False):
        return self.model.data(row, col, skip_format=skip_format)

    def grid_data_import(self):
        file, filetype = q2app.q2_app.get_open_file_dialoq(
            q2app.MESSAGE_GRID_DATA_IMPORT_TITLE, filter="CSV (*.csv);;JSON(*.json)"
        )
        if not file:
            return
        file = self.validate_impexp_file_name(file, filetype)
        waitbar = self._q2dialogs.Q2WaitShow(q2app.MESSAGE_GRID_DATA_IMPORT_WAIT % file)
        try:
            self.model.data_import(file, tick_callback=lambda: waitbar.step(1000))
        except Exception as e:
            self._q2dialogs.q2Mess(
                q2app.MESSAGE_GRID_DATA_IMPORT_ERROR % (self.db.last_sql_error + ": " + str(e))
            )
            waitbar.close()
        else:
            _count, _time = waitbar.close()
            self._q2dialogs.q2Mess(q2app.MESSAGE_GRID_DATA_IMPORT_DONE % (_count, _time))

    def set_grid_row_colors(self):
        selected_rows = self.get_grid_selected_rows()
        if selected_rows:
            _form: Q2Form = self.q2_app.Q2Form(q2app.ACTION_TOOLS_COLOR_TEXT)
            _form.add_control("/v")
            q2_bcolor = self.r.__getattr__("q2_bcolor")
            _bcolor = f"#{int_(q2_bcolor):06x}"
            q2_fcolor = self.r.__getattr__("q2_fcolor")
            _fcolor = f"#{int_(q2_fcolor):06x}"
            _form.add_control("fcolor", _("Font color"), control="color", data=_fcolor, datalen=20)
            _form.add_control("bcolor", _("Background color"), control="color", data=_bcolor, datalen=20)

            def reset_colors():
                _form.s.fcolor = "#000000"
                _form.s.bcolor = "#000000"

            _form.add_control(
                "reset_colors", _("Reset colors"), datalen=15, control="button", valid=reset_colors
            )
            _form.ok_button = True
            _form.cancel_button = True

            def colors_valid():
                not_valid = False
                color = _form.s.fcolor
                _form.s.fcolor = "#000000"
                _form.s.fcolor = color
                if color != _form.s.fcolor:
                    self._q2dialogs.q2Mess(_("Foreground color is not valid! "))
                    not_valid = True

                color = _form.s.bcolor
                _form.s.bcolor = "#000000"
                _form.s.bcolor = color
                if color != _form.s.bcolor:
                    self._q2dialogs.q2Mess(_("Background color is not valid! "))
                    not_valid = True

                return not not_valid

            _form.valid = colors_valid
            _form.run()
            if _form.ok_pressed:
                pk = self.model.get_meta_primary_key()
                waitbar = None
                if len(selected_rows) > 10:
                    waitbar = self.show_progressbar(q2app.MESSAGE_ROWS_COLOR, len(selected_rows))
                for row in selected_rows:
                    if waitbar:
                        waitbar.step(1000)
                    dic = {pk: self.model.get_record(row)[pk]}
                    _form.s.fcolor = _form.s.fcolor if _form.s.fcolor else "0"
                    _form.s.bcolor = _form.s.bcolor if _form.s.bcolor else "0"

                    dic["q2_fcolor"] = int(_form.s.fcolor.lstrip("#"), 16)
                    dic["q2_bcolor"] = int(_form.s.bcolor.lstrip("#"), 16)
                    self.model.update(dic, self.current_row, refresh=False)
                self.model.refresh()
                self.set_grid_index(self.current_row)
                if waitbar:
                    waitbar.close()

    def grid_data_paste_clipboard(self):
        Q2PasteClipboard(self)

    def grid_data_bulk_update(self):
        Q2BulkUpdate(self)

    def grid_print(self):
        self._q2dialogs.q2mess("Must be implemented")

    def grid_data_info(self):
        columns = self.model.columns
        self._q2dialogs.q2Mess(
            f"{q2app.GRID_DATA_INFO_TABLE}: {self.model.get_table_name()}"
            f"<br>{q2app.GRID_DATA_INFO_ROWS}: {self.model.row_count()}"
            f"<br>{q2app.GRID_DATA_INFO_ORDER}: {self.model.get_order()}"
            f"<br>{q2app.GRID_DATA_INFO_FILTER}: {html.escape(self.model.get_where())}"
            f"<br>{q2app.GRID_DATA_INFO_COLUMNS}: {columns}"
        )

    def set_style_sheet(self, css: str):
        self.style_sheet = css
        for x in self.form_stack:
            x.set_style_sheet(self.style_sheet)

    def prepare_where(self, column=None, control1=None, control2=None):
        mem_widgets = self.widgets().keys()
        if control1 is None:
            if column in mem_widgets:
                control1 = column
            elif column + "____1" in mem_widgets:
                control1 = column + "____1"
        if control1 not in mem_widgets:
            return ""

        if not self.w.__getattr__(control1).is_checked():
            return ""

        # control_datatype = self.controls.c.__getattr__(control1)["datatype"]
        date_control = self.controls.c.__getattr__(control1)["datatype"] == "date"
        num_control = self.controls.c.__getattr__(control1).get("num")
        control1_value = self.s.__getattr__(control1)
        if control2 is None:
            if control1.endswith("____1"):
                control2 = control1[:-5] + "____2"
                control2_value = self.s.__getattr__(control2)
            else:
                control2_value = None
        if date_control:
            if control1_value == "0000-00-00":
                control1_value = ""
            if control2_value == "0000-00-00":
                control2_value = ""
        elif num_control:
            control1_value = num(control1_value)
            if control2_value:
                control2_value = num(control2_value)

        if (control2_value is None) or control1_value == control2_value:
            if date_control or num_control:
                return f"{column} = '{control1_value}'"
            else:
                return f"{column} like '%{control1_value}%'"
        elif (control1_value and not control2_value) or (control2_value and control1_value > control2_value):
            return f"{column} >= '{control1_value}'"
        elif not control1_value and control2_value:
            return f"{column} <= '{control2_value}'"
        elif control1_value and control2_value:
            return f"{column} >= '{control1_value}' and {column}<='{control2_value}'"
        return ""


class Q2FormWindow:
    def __init__(self, q2_form: Q2Form, title=""):
        super().__init__()
        self.title = title if title else q2_form.title
        self.shown = False
        self.q2_form = q2_form
        self.form_is_active = False
        # self.title = ""
        self.widgets = {}
        self.tab_widget_list = []
        self.tab_widget = None
        # Must be defined in any subclass
        self._widgets_package = None
        self.escape_enabled = True
        self.mode = "form"
        self.hotkey_widgets = {}
        self.grid_actions = q2app.Q2Actions()
        self._in_close_flag = None
        self.is_crud = None
        self.not_closed = True

    def on_activate(self):
        if self.mode == "grid":
            self.q2_form.q2_app.show_statusbar_mess(self.q2_form.model.row_count())

    def create_grid_navigation_actions(self):
        """returns standard actions for the grid"""
        actions = q2app.Q2Actions()
        actions.add_action(text="-")
        actions.add_action(
            text=q2app.ACTION_FIRST_ROW_TEXT,
            worker=lambda: self.move_grid_index(7),
            icon=q2app.ACTION_FIRST_ROW_ICON,
            hotkey=q2app.ACTION_FIRST_ROW_HOTKEY,
            eof_disabled=1,
        )
        actions.add_action(
            text=q2app.ACTION_PREVIOUS_ROW_TEXT,
            worker=lambda: self.move_grid_index(8),
            icon=q2app.ACTION_PREVIOUS_ROW_ICON,
            eof_disabled=1,
        )
        actions.add_action(
            text=q2app.ACTION_REFRESH_TEXT,
            worker=lambda: self.q2_form.refresh(),
            icon=q2app.ACTION_REFRESH_ICON,
            hotkey=q2app.ACTION_REFRESH_HOTKEY,
        )
        actions.add_action(
            text=q2app.ACTION_NEXT_ROW_TEXT,
            worker=lambda: self.move_grid_index(2),
            icon=q2app.ACTION_NEXT_ROW_ICON,
            eof_disabled=1,
        )
        actions.add_action(
            text=q2app.ACTION_LAST_ROW_TEXT,
            worker=lambda: self.move_grid_index(1),
            icon=q2app.ACTION_LAST_ROW_ICON,
            hotkey=q2app.ACTION_LAST_ROW_HOTKEY,
            eof_disabled=1,
        )
        # HIIDEN rows menu

        show_color_menu = (
            hasattr(self.q2_form.model, "check_db_column")
            and self.q2_form.model.check_db_column("q2_bcolor")
            and not self.q2_form.model.readonly
        )

        show_hide_menu = (
            hasattr(self.q2_form.model, "check_db_column")
            and self.q2_form.model.check_db_column("q2_hidden")
            and not self.q2_form.model.readonly
        )

        if show_color_menu or show_hide_menu:
            actions.add_action(text="-")
            actions.add_action(
                text=q2app.ACTION_HIDDEN_ROW_TEXT,
                icon=q2app.ACTION_HIDDEN_ROW_ICON,
            )
        if show_color_menu:
            actions.add_action(
                text=q2app.ACTION_HIDDEN_ROW_TEXT + "|" + q2app.ACTION_TOOLS_COLOR_TEXT,
                icon=q2app.ACTION_TOOLS_COLOR_ICON,
                worker=self.q2_form.set_grid_row_colors,
                eof_disabled=1,
            )
            actions.add_action(
                text=q2app.ACTION_HIDDEN_ROW_TEXT + "|-",
            )
        if show_hide_menu:
            actions.add_action(
                text=q2app.ACTION_HIDDEN_ROW_TEXT + "|" + q2app.ACTION_HIDE_ROW_TEXT,
                icon=q2app.ACTION_HIDE_ROW_ICON,
                worker=self.q2_form.hidden_row_hide,
                eof_disabled=1,
            )
            actions.add_action(
                text=q2app.ACTION_HIDDEN_ROW_TEXT + "|" + q2app.ACTION_SHOW_ROW_TEXT,
                icon=q2app.ACTION_SHOW_ROW_ICON,
                worker=self.q2_form.hidden_row_show,
                eof_disabled=1,
            )
            actions.add_action(
                text=q2app.ACTION_HIDDEN_ROW_TEXT + "|-",
            )
            actions.add_action(
                text=q2app.ACTION_HIDDEN_ROW_TEXT + "|" + q2app.ACTION_SHOW_NOTHIDDEN_TEXT,
                icon=q2app.ACTION_SHOW_NOTHIDDEN_ICON,
                worker=self.q2_form.hidden_row_show_not_hidden,
            )
            actions.add_action(
                text=q2app.ACTION_HIDDEN_ROW_TEXT + "|" + q2app.ACTION_SHOW_ALL_TEXT,
                icon=q2app.ACTION_SHOW_ALL_ICON,
                worker=self.q2_form.hidden_row_show_all,
            )
            actions.add_action(
                text=q2app.ACTION_HIDDEN_ROW_TEXT + "|" + q2app.ACTION_SHOW_HIDDEN_TEXT,
                icon=q2app.ACTION_SHOW_HIDDEN_ICON,
                worker=self.q2_form.hidden_row_show_hidden,
            )

        actions.add_action(text="-")
        actions.add_action(
            text=q2app.ACTION_TOOLS_TEXT,
            icon=q2app.ACTION_TOOLS_ICON,
        )
        actions.add_action(
            text=q2app.ACTION_TOOLS_TEXT + "|" + q2app.ACTION_TOOLS_EXPORT_TEXT,
            worker=self.q2_form.grid_data_export,
            icon=q2app.ACTION_TOOLS_EXPORT_ICON,
            eof_disabled=1,
        )

        if self.q2_form.is_grid_updateable():
            actions.add_action(
                text=q2app.ACTION_TOOLS_TEXT + "|" + q2app.ACTION_TOOLS_IMPORT_TEXT,
                worker=self.q2_form.grid_data_import,
                icon=q2app.ACTION_TOOLS_IMPORT_ICON,
            )
        actions.add_action(
            text=q2app.ACTION_TOOLS_TEXT + "|-",
        )

        actions.add_action(
            text=q2app.ACTION_TOOLS_TEXT + "|" + q2app.ACTION_TOOLS_IMPORT_CLIPBOARD_TEXT,
            worker=self.q2_form.grid_data_paste_clipboard,
            icon=q2app.ACTION_TOOLS_IMPORT_CLIPBOARD_ICON,
        )

        actions.add_action(
            text=q2app.ACTION_TOOLS_TEXT + "|" + q2app.ACTION_TOOLS_BULK_UPDATE_TEXT,
            worker=self.q2_form.grid_data_bulk_update,
            icon=q2app.ACTION_TOOLS_BULK_UPDATE_ICON,
            eof_disabled=1,
        )

        actions.add_action(
            text=q2app.ACTION_TOOLS_TEXT + "|-",
        )

        actions.add_action(
            text=q2app.ACTION_TOOLS_TEXT + "|" + "Print",
            worker=self.q2_form.grid_print,
            icon="print",
            eof_disabled=1,
        )

        actions.add_action(
            text=q2app.ACTION_TOOLS_TEXT + "|-",
        )

        actions.add_action(
            text=q2app.ACTION_TOOLS_TEXT + "|" + q2app.ACTION_TOOLS_INFO_TEXT,
            worker=self.q2_form.grid_data_info,
            icon=q2app.ACTION_TOOLS_INFO_ICON,
        )

        if not self.q2_form.i_am_child:
            actions.add_action(text="-")
            actions.add_action(
                text=q2app.ACTION_CLOSE_TEXT, worker=self.close, icon=q2app.ACTION_CLOSE_ICON, tag="orange"
            )
        self.q2_form.grid_navigation_actions_hook(actions)
        return actions

    def move_grid_index(self, direction=None):
        """Directions - look at numpad to get the idea"""
        if direction == 7:  # Top
            self.set_grid_index(0, self.get_grid_index()[1])
        elif direction == 8:  # Up
            self.set_grid_index(self.get_grid_index()[0] - 1, self.get_grid_index()[1])
        elif direction == 2:  # Down
            self.set_grid_index(self.get_grid_index()[0] + 1, self.get_grid_index()[1])
        elif direction == 1:  # Last
            self.set_grid_index(self.get_grid_row_count(), self.get_grid_index()[1])

    def set_grid_index(self, row=0, col=0):
        self.widgets["form__grid"].set_index(row, col)

    def get_grid_index(self):
        return self.widgets["form__grid"].current_index()

    def get_grid_selected_rows(self):
        return self.widgets["form__grid"].get_selected_rows()

    def set_grid_selected_rows(self, index_list=[]):
        self.widgets["form__grid"].set_selected_rows(index_list)

    def get_grid_row_count(self):
        return self.widgets["form__grid"].row_count()

    def build_grid(self):
        # populate model with columns metadata
        self.mode = "grid"
        tmp_grid_form = Q2Form()
        tmp_grid_form.add_control("/vs", tag="gridsplitter")
        self.q2_form.grid_navi_actions = self.create_grid_navigation_actions()

        tmp_grid_form.add_control(
            "form__grid",
            control="grid",
            actions=[self.q2_form.actions, self.q2_form.grid_navi_actions],
            # stretch=100,
        )
        # place child forms
        if self.q2_form.max_child_level:
            for action in self.q2_form.actions:
                if action.get("child_form") and not action.get("child_noshow"):
                    tmp_grid_form.add_control("/t", action.get("text", "="), stretch=1)
                    #  create child form!
                    action["child_form_object"] = action.get("child_form")()
                    action["child_form_object"].prev_form = self.q2_form
                    action["child_form_object"].title = self.q2_form.title + " / " + action["text"]
                    action["child_form_object"].i_am_child = True
                    action["child_form_object"].max_child_level = self.q2_form.max_child_level - 1
                    self.q2_form.children_forms.append(action)
                    tmp_grid_form.add_control(
                        f"child_grid__{action['text']}",
                        widget=action["child_form_object"],
                    )
        tmp_grid_form.add_control("/")

        if self.q2_form.show_app_modal_form is False:
            tmp_grid_form.controls[-1], tmp_grid_form.controls[-2] = (
                tmp_grid_form.controls[-2],
                tmp_grid_form.controls[-1],
            )
        self.build_form(tmp_grid_form.get_controls())
        self.q2_form.refresh_children()
        self.move_grid_index(1)

    def grid_navigation_actions_hook(self, actions):
        pass

    def build_form(self, controls=[]):
        frame_stack = [self]
        tmp_frame = None

        if controls == []:
            controls = self.q2_form.get_controls()
        # set deafault layout to Form if first line not a layout def
        if controls and not controls[0].get("column", "").startswith("/"):
            controls.insert(0, {"column": "/f"})
        # Create widgets
        for meta in controls:
            meta["form"] = self.q2_form
            meta["q2_app"] = q2app.q2_app
            meta["form_window"] = self
            if meta.get("noform", ""):
                continue
            meta = q2app.Q2Controls.validate(meta)
            current_frame = frame_stack[-1]
            # do not add widget if it is not first tabpage on the form
            if not (meta.get("column", "") == ("/t") and self.tab_widget is not None):
                # get widgets to add
                label2add, widget2add = self.widget(meta)

                if current_frame.frame_mode == "f":  # form layout
                    # if label2add is not None:
                    #     label2add.set_content_margins(10, int(q2app.q2_app.get_char_height() / 4), 2, 0)
                    if hasattr(widget2add, "frame_mode") and not meta.get("relation"):
                        # add any frame into form frame
                        if label2add is None:
                            label2add = self._get_widget("label")({"label": meta.get("label", "")})
                            widget2add.label = label2add
                        widget2add.hide_border()
                    current_frame.add_row(label2add, widget2add)
                else:  # v- h- box layout
                    if label2add is not None:
                        if (
                            current_frame != self
                            and current_frame.get_widget_count() == 0
                            and current_frame.label
                            and frame_stack[-2].frame_mode == "f"
                        ):
                            # when frame prev frame is FORM and it is first widget
                            # in the current frame - move widget label to frame label
                            # which is prev FORM frame )
                            current_frame.label.set_text(label2add.get_text())
                            if widget2add:
                                widget2add.label = current_frame.label
                        else:
                            current_frame.add_widget(label2add)
                    if widget2add is not None:
                        if meta.get("column", "") in ("/vr", "/hr"):  # scroller
                            scroller = self._get_widget("scroller")({"widget": widget2add})
                            current_frame.add_widget(scroller)
                        else:
                            current_frame.add_widget(widget2add)
                        if meta.get("control") == "toolbar":  # context menu for frame
                            widget2add.set_context_menu(current_frame)
            # Hotkeys
            if meta.get("hotkey") and meta.get("valid"):
                if meta.get("hotkey") not in self.hotkey_widgets:
                    self.hotkey_widgets[meta.get("hotkey")] = []
                self.hotkey_widgets[meta.get("hotkey")].append(widget2add)
            # Special cases
            if meta.get("column", "") == ("/t"):
                if self.tab_widget is None:
                    self.tab_widget = widget2add
                    frame_stack.append(widget2add)
                    self.tab_widget_list.append(widget2add)
                else:  # If second and more tabpage widget
                    if tmp_frame in frame_stack:
                        frame_stack = frame_stack[: frame_stack.index(tmp_frame)]
                tmp_frame = self.widget({"column": "/v"})[1]
                self.tab_widget.add_tab(tmp_frame, meta.get("label", ""))
                frame_stack.append(tmp_frame)
            elif meta.get("column", "") == ("/s"):
                continue  # do not touch - see elif +2
            elif meta.get("column", "") == "/":
                if len(frame_stack) > 1:
                    frame_stack.pop()
                    # Remove tab widget if it is at the end of stack
                    if "q2tab.q2tab" in f"{type(frame_stack[-1])}":
                        self.tab_widget = None
                        frame_stack.pop()
            elif meta.get("column", "").startswith("/"):
                frame_stack.append(widget2add)

        if len(self.tab_widget_list) > 1:
            for x in self.tab_widget_list:
                x.set_shortcuts_local()

        # Make it no more working
        self.build_grid = lambda: None
        self.build_form = lambda: None

    def widget(self, meta):
        """Widgets fabric"""
        if not meta.get("control") or meta.get("control") == "":
            if meta.get("widget"):
                control = "widget"
            else:
                control = "line" if meta.get("column") else "label"
        else:
            control = meta.get("control")

        if meta.get("to_table"):  # relation is here
            control = "relation"

        if control == "":
            control = "label"

        column = meta.get("column", "")
        label = meta.get("label", "")
        class_name = ""

        widget2add = None
        if label and control not in NO_LABEL_WIDGETS:
            label2add = self._get_widget("label")(meta)
        else:
            label2add = None

        # Form or widget
        if control == "widget":
            if isinstance(meta.get("widget"), Q2Form):
                if meta.get("widget").model is not None:
                    widget2add = meta.get("widget").get_grid_widget()
                else:
                    widget2add = meta.get("widget").get_form_widget()
                widget2add.meta = meta
                widget2add.form_is_active = True
            else:
                widget2add = meta.get("widget")
            if not hasattr(widget2add, "meta"):
                setattr(widget2add, "meta", meta)
        else:  # Special cases
            if column[:2] in ("/h", "/v", "/f"):  # frame
                control = "frame"
                class_name = "frame"
                label2add = None
            elif "/" == column:
                return None, None
            elif "/t" in column:  # Tabpage
                label2add = None
                control = "tab"
            elif control.startswith("code"):
                control = "code"
            elif "radio" in control:
                control = "radio"
            elif "toolbar" in control:
                control = "toolbar"
            elif column == "/s":
                control = "space"

            widget_class = self._get_widget(control, class_name)
            widget2add = widget_class(meta)

            if hasattr(widget2add, "label"):
                widget2add.label = label2add
        if meta.get("check"):  # has checkbox
            # label2add = self._get_widget("check", "check")({"label": meta["label"], "stretch": 0})
            label2add = self._get_widget("check", "check")(
                {"label": meta["label"] if meta["control"] != "check" else "Turn on", "stretch": 0}
            )

            label2add.add_managed_widget(widget2add)
            if not meta.get("data"):
                widget2add.set_disabled()
            else:
                label2add.set_checked()

        self.widgets[meta.get("tag", "") if meta.get("tag", "") else column] = widget2add

        action2add = None
        if meta.get("actions") and meta.get("control") != "toolbar":
            action2add = self._get_widget("toolbar", "toolbar")(
                {
                    "control": "toolbar",
                    "actions": meta["actions"],
                    "form": self.q2_form,
                    "stretch": 0,
                }
            )
            action2add.set_context_menu(widget2add)
            action2add.fix_default_height()
            # Actions!
            if widget2add and action2add:
                if control == "label" or (not action2add.show_main_button and not action2add.show_actions):
                    action_frame = self._get_widget("frame")({"column": "/v"})
                else:
                    #  Splitter!
                    action_frame = self._get_widget("frame")({"column": "/vs"})
                    self.widgets[meta.get("tag", "") if meta.get("tag", "") else column + "_splitter"] = (
                        action_frame
                    )
                action_frame.add_widget(action2add)
                action_frame.add_widget(widget2add)
                widget2add = action_frame

        return label2add, widget2add

    def _get_widget(self, module_name, class_name=""):
        """For given name returns class from current GUI engine module"""
        if class_name == "":
            class_name = module_name
        module_name = f"q2{module_name}"
        class_name = f"q2{class_name}"
        try:
            # print(self._widgets_package, module_name)
            return getattr(getattr(self._widgets_package, module_name), class_name)
        except Exception:
            # print(self._widgets_package, module_name, class_name)
            return getattr(getattr(self._widgets_package, "q2label"), "q2label")

    def show_form(self, modal="modal", no_build=False):
        if no_build is False:
            self.build_form()
        self.set_style_sheet(self.q2_form.style_sheet)

        self.q2_form.form_stack.append(self)

        # Restore grid columns sizes
        self.restore_splitters()
        self.restore_grid_columns()

        if self.mode == "grid":
            if self.q2_form.before_grid_show() is False:
                self.q2_form.form_stack.pop()
                return
        elif self.mode == "form":
            self.form_is_active = True
            if self.q2_form.before_form_show() is False:
                self.q2_form.form_stack.pop()
                return
        # search for the first enabled widget
        for x, widget in self.q2_form.widgets().items():
            if not x.startswith("/") and widget is not None:
                if hasattr(widget, "can_get_focus") and not widget.can_get_focus():
                    continue
                elif hasattr(widget, "is_enabled") and widget.is_enabled():
                    widget.set_focus()
                    break
                elif hasattr(widget, "get_check") and widget.get_check():
                    if widget.get_check().is_enabled():
                        widget.get_check().set_focus()
                    break
        self.q2_form.q2_app.show_form(self, modal)
        # print(">>", len(set(self.q2_form.q2_app.QApplication.allWidgets())))

    def get_controls_list(self, name: str):
        return [self.widgets[x] for x in self.widgets if type(self.widgets[x]).__name__ == name]

    def restore_splitters(self):
        # Restore splitters sizes
        for x in self.get_splitters():
            sizes = q2app.q2_app.settings.get(
                self.window_title,
                f"splitter-{x}",
                "",
            )
            self.widgets[x].splitter.set_sizes(sizes)

    def restore_grid_columns(self):
        # for grid in self.get_grid_list():
        for grid in self.get_controls_list("q2grid"):
            col_settings = {}
            for count, x in enumerate(self.q2_form.model.headers):
                data = q2app.q2_app.settings.get(self.window_title, f"grid_column__'{x}'")
                if data == "":
                    if (
                        self.q2_form.model.meta[count].get("relation")
                        or self.q2_form.model.meta[count].get("num") is None
                    ):
                        c_w = q2app.GRID_COLUMN_WIDTH
                    else:
                        c_w = int_(self.q2_form.model.meta[count].get("datalen"))
                    c_w = int(q2app.q2_app.get_char_width() * (min(c_w, q2app.GRID_COLUMN_WIDTH)))
                    data = f"{count}, {c_w}"
                if len(self.q2_form.model.headers) == 1:
                    data = "0, 3000"
                col_settings[x] = data
            grid.set_column_settings(col_settings)
        for x in self.get_controls_list("Q2FormWindow"):
            x.restore_grid_columns()
            x.restore_splitters()

    def save_grid_columns(self):
        for grid in self.get_controls_list("q2grid"):
            for x in grid.get_columns_settings():
                q2app.q2_app.settings.set(
                    self.window_title,
                    f"grid_column__'{x['name']}'",
                    x["data"],
                )
        for x in self.get_controls_list("Q2FormWindow"):
            x.close()

    def close(self):
        if self._in_close_flag:
            return
        self._in_close_flag = True
        if self in self.q2_form.form_stack[-1:]:
            self.q2_form.form_stack.pop()
        self.save_splitters()
        self.save_grid_columns()
        if hasattr(q2app.q2_app, "settings"):
            self.save_geometry(q2app.q2_app.settings)
        self.q2_form._close(self)
        if self.is_crud:
            self.q2_form.crud_form = None

    def save_splitters(self):
        if not hasattr(q2app.q2_app, "settings"):
            return
        for x in self.get_splitters():
            q2app.q2_app.settings.set(
                self.window_title,
                f"splitter-{x}",
                self.widgets[x].splitter.get_sizes(),
            )

    def get_splitters(self):
        return [
            x
            for x in self.widgets.keys()
            if hasattr(self.widgets[x], "splitter") and self.widgets[x].splitter is not None
        ]

    def set_style_sheet(self, css):
        pass


class Q2FormData:
    """Get and put data from/to form"""

    def __init__(self, q2_form: Q2Form):
        self.q2_form = q2_form

    def __setattr__(self, name, value):
        if name != "q2_form":
            if self.q2_form.crud_form:
                widget = self.q2_form.crud_form.widgets.get(name)
            elif self.q2_form.form_stack:
                widget = self.q2_form.form_stack[-1].widgets.get(name)
            else:
                widget = None
            if hasattr(widget, "set_text"):
                widget.set_text(value)
            else:  # no widget - put data to model's record
                self.q2_form._model_record[name] = value
        else:
            self.__dict__[name] = value

    def __getattr__(self, name):
        if self.q2_form.crud_form:
            widget = self.q2_form.crud_form.widgets.get(name)
        elif self.q2_form.form_stack == []:
            if self.q2_form.last_closed_form is None:
                return None
            else:
                # widget = self.q2_form.last_closed_form.widgets.get(name)
                # widget = self.q2_form.last_closed_form_widgets_text.get(name)
                return self.q2_form.last_closed_form_widgets_text.get(name, "")
        else:
            widget = self.q2_form.form_stack[-1].widgets.get(name)
        if widget is not None:
            if hasattr(widget, "get_text"):
                return widget.get_text()
            else:
                return ""
        else:  # no widget here? get data from model
            return self.q2_form._model_record.get(name, None)


class Q2FormWidget:
    """Get widget object from form"""

    def __init__(self, q2_form: Q2Form):
        self.q2_form = q2_form

    def __getattr__(self, attrname):
        widget = None
        if self.q2_form.crud_form:
            widgets = self.q2_form.crud_form.widgets
        elif self.q2_form.form_stack == []:
            if self.q2_form.last_closed_form is None:
                return None
            else:
                widgets = self.q2_form.last_closed_form.widgets
                widgets = self.last_closed_form_widgets
        else:
            widgets = self.q2_form.form_stack[-1].widgets
        if attrname.startswith("_") and attrname.endswith("_"):
            pos = int_(attrname.replace("_", ""))
            if pos < len(widgets):
                widget = widgets.get(list(widgets)[pos])
        else:
            widget = widgets.get(attrname)
        if widget is not None:
            return widget


class Q2FormAction:
    def __init__(self, q2_form):
        self.q2_form: Q2Form = q2_form

    def tag(self, tag=""):
        if tag:
            for act in self.q2_form.actions:
                if act.get("tag") == tag:
                    return act
        return {}

    def __getattr__(self, name):
        for act in self.q2_form.actions:
            if act.get("text") == name:
                return act
        return {}


class Q2ModelData:
    def __init__(self, q2_form: Q2Form):
        self.q2_form = q2_form

    def __getattr__(self, name):
        self.q2_form.model.refresh_record(self.q2_form.current_row)
        datadic = self.q2_form.model.get_record(self.q2_form.current_row)
        return datadic.get(name, "")


class Q2PasteClipboard:
    def __init__(self, q2_form: Q2Form):
        self.q2_form = q2_form
        self.cliptext = q2app.q2_app.get_clipboard_text()
        self.clipboard_headers = self.cliptext.split("\n")[0].split("\t")
        self.load_target()
        self.load_source()
        self.load_data()
        self.load_set()
        if self.show_main_form().ok_pressed:
            self.save_set()
            self.paste_to_form()

    def paste_to_form(self):
        source_map = {x["source_column"]: x["_target_column"] for x in self.target_data if x["source_column"]}
        if not self.main_form.s.first_row_is_header:
            self.data_data.insert(0, {f"{x}": x for x in self.clipboard_headers})

        waitbar = self.q2_form.show_progressbar(q2app.PASTE_CLIPBOARD_WAIT, len(self.data_data))

        self.q2_form.show_crud_form(NEW, modal="")

        for row in self.data_data:
            waitbar.step(1000)
            for col in row:
                if col in source_map:
                    col_def = self.q2_form.c.__getattr__(source_map[col])
                    if col_def.get("num") and int_(col_def.get("datadec")) > 0 and "," in row[col]:
                        row[col] = nums(row[col])
                    self.q2_form.s.__setattr__(source_map[col], row[col])
            self.q2_form.crud_save(close_form=False)
            self.q2_form.before_form_show()
        self.q2_form.close()
        waitbar.close()

    def save_set(self):
        q2app.q2_app.settings.set(
            self.q2_form.title,
            f"paste-{self.target_hash}-{self.source_hash}",
            json.dumps(self.target_data),
        )

    def show_main_form(self):
        self.main_form = self.q2_form.q2_app.Q2Form(q2app.PASTE_CLIPBOARD_TITLE)
        self.main_form.add_control("/v")
        self.main_form.add_control(
            "first_row_is_header", q2app.PASTE_CLIPBOARD_FIRST_ROW, control="check", data="*"
        )
        self.main_form.add_control("/vs", tag="vs")
        self.main_form.add_control("/hs", tag="hs")
        self.main_form.add_control("target_form", widget=self.target_form)

        # csv_form.add_control("/v")
        # csv_form.add_control("/s")
        # csv_form.add_control("to_target", "<>", datalen=4, control="button", valid=self.move_it)
        # csv_form.add_control("/")

        self.main_form.add_control("source_form", widget=self.source_form)
        self.main_form.add_control("/")
        self.main_form.add_control("", q2app.PASTE_CLIPBOARD_CLIPBOARD_DATA)
        self.main_form.add_control("data_form", widget=self.data_form)

        self.main_form.cancel_button = True
        self.main_form.ok_button = True
        self.main_form.add_ok_cancel_buttons()
        self.main_form.show_form()

        return self.main_form

    def move_it(self):
        target_row = self.target_form.current_row
        source_row = self.source_form.current_row
        target_row_data = self.target_form.get_current_record()
        source_row_data = self.source_form.get_current_record()
        if target_row_data["source_column"] == "" and source_row_data["column"]:
            # move left
            target_row_data["source_column"] = source_row_data["column"]
            source_row_data["column"] = ""
        else:
            # swap
            target_row_data["source_column"], source_row_data["column"] = (
                source_row_data["column"],
                target_row_data["source_column"],
            )

        self.target_form.model.update(target_row_data, target_row)
        self.source_form.model.update(source_row_data, source_row)

        if self.target_form.w.form__grid.has_focus():
            source_row += 1
        elif self.source_form.w.form__grid.has_focus():
            target_row += 1
        self.target_form.set_grid_index(target_row)
        self.source_form.set_grid_index(source_row)

    def load_set(self):
        last_set = q2app.q2_app.settings.get(
            self.q2_form.title,
            f"paste-{self.target_hash}-{self.source_hash}",
            None,
        )

        if last_set:
            self.target_data = json.loads(last_set)
            self.target_form.model.set_records(self.target_data)
            for x in self.target_data:
                col_name = x["source_column"]
                for row, dic in enumerate(self.source_data):
                    if dic["column"] == col_name:
                        self.source_data[row]["column"] = ""
                        break

    def myhash(self, mystr):
        rez = 1
        for x in range(len(mystr)):
            rez += rez * x * ord(mystr[x])
        return rez & 0xFFFFFFFF

    def load_target(self):
        self.target_form = self.q2_form.q2_app.Q2Form(q2app.PASTE_CLIPBOARD_TARGET)
        self.target_data = []
        target_hash_string = ""
        for x in self.q2_form.controls:
            if x["migrate"]:
                target_hash_string += x.get("column")
                self.target_data.append(
                    {
                        "target_column": f'{x.get("label") if x.get("label") else x.get("gridlabel")} '
                        + f'({self.q2_form.model.get_table_name()}.{x.get("column")})',
                        "_target_column": x.get("column"),
                        "source_column": "",
                    },
                )
        self.target_hash = self.myhash(target_hash_string)

        self.target_form.set_model(Q2Model())
        self.target_form.model.set_records(self.target_data)
        self.target_form.add_control(
            "target_column", q2app.PASTE_CLIPBOARD_TARGET_COLUMNS, control="line", datalen=100
        )
        self.target_form.add_control(
            "source_column", q2app.PASTE_CLIPBOARD_SOURCE_COLUMNS, control="line", datalen=100
        )
        self.target_form.grid_double_clicked = self.move_it
        self.target_form.i_am_child = 1

    def load_source(self):
        self.source_data = [{"column": x} for x in self.clipboard_headers]
        self.source_hash = self.myhash(",".join(self.clipboard_headers))
        self.source_form = self.q2_form.q2_app.Q2Form(q2app.PASTE_CLIPBOARD_SOURCE)

        self.source_form.set_model(Q2Model())
        self.source_form.model.set_records(self.source_data)
        self.source_form.add_control(
            "column", q2app.PASTE_CLIPBOARD_SOURCE_COLUMNS, control="line", datalen=100
        )
        self.source_form.grid_double_clicked = self.move_it
        self.source_form.i_am_child = 1

    def load_data(self):
        self.data_data = []
        for cliptext_row in self.cliptext.split("\n")[1:]:
            row_dic = {}
            for ncol, cliptext_column in enumerate(cliptext_row.split("\t")):
                row_dic[self.clipboard_headers[ncol]] = cliptext_column
            if len(self.clipboard_headers) != len(row_dic):
                continue
            self.data_data.append(row_dic)

        self.data_form = self.q2_form.q2_app.Q2Form(q2app.PASTE_CLIPBOARD_CLIPBOARD_DATA)
        for col in self.clipboard_headers:
            self.data_form.add_control(col, col)

        self.data_form.set_model(Q2Model())
        self.data_form.model.set_records(self.data_data)
        self.data_form.i_am_child = 1


class Q2BulkUpdate:
    def __init__(self, q2_form: Q2Form):
        self.q2_form = q2_form

        self.load_target()
        if self.show_main_form().ok_pressed:
            self.bulk_data_enter()

    def bulk_data_enter(self):
        bulk_data_form = self.q2_form.q2_app.Q2Form(q2app.BULK_DATA_ENTRY_TITLE)
        bulk_data_form.model = self.q2_form.model
        bulk_columns = []
        current_record = self.q2_form.get_current_record()
        for x in self.target_data:
            if x["_selected"]:
                control = dict(x)
                control["data"] = current_record.get(control["column"], "")
                bulk_data_form.add_control(**control)
                bulk_columns.append(control["column"])
        if len(bulk_data_form.controls) == 0:
            return

        bulk_data_form.ok_button = True
        bulk_data_form.cancel_button = True
        bulk_data_form.show_form()
        if bulk_data_form.ok_pressed:
            self.bulk_update(bulk_data_form, bulk_columns)

    def bulk_update(self, bulk_data_form, bulk_columns):
        record_list = []
        for x in self.q2_form.get_grid_selected_rows():
            record_list.append(self.q2_form.model.get_record(x))
        waitbar = self.q2_form.show_progressbar(q2app.BULK_DATA_WAIT, len(record_list))
        for x in record_list:
            waitbar.step(1000)
            self.q2_form.set_grid_index(self.q2_form.model.seek_row(x))
            self.q2_form.show_crud_form(EDIT, modal="")
            for bulk_column in bulk_columns:
                self.q2_form.w.__getattr__(bulk_column).when()
                self.q2_form.s.__setattr__(bulk_column, bulk_data_form.s.__getattr__(bulk_column))
                self.q2_form.w.__getattr__(bulk_column).valid()
            self.q2_form.crud_save()
        waitbar.close()

    def show_main_form(self):
        self.main_form = self.q2_form.q2_app.Q2Form(q2app.BULK_DATA_MAIN_TITLE)
        self.main_form.add_control("/v")
        self.main_form.add_control("target_form", widget=self.target_form)
        self.main_form.cancel_button = True
        self.main_form.ok_button = True
        self.main_form.add_ok_cancel_buttons()
        self.main_form.show_form()
        return self.main_form

    def select(self):
        target_row = self.target_form.current_row
        target_row_data = self.target_form.get_current_record()
        target_row_data["_selected"] = "" if target_row_data["_selected"] else "*"
        self.target_form.model.update(target_row_data, target_row)
        self.target_form.set_grid_index(target_row)

    def load_target(self):
        self.target_form = self.q2_form.q2_app.Q2Form(q2app.BULK_TARGET_TITLE)
        self.target_data = []
        for x in self.q2_form.controls:
            if not x["pk"] and not x["noform"] and not x["column"].startswith("/"):
                x["target_column"] = (
                    f'{x.get("label") if x.get("label") else x.get("gridlabel")} '
                    + f'({self.q2_form.model.get_table_name()}.{x.get("column")})'
                )
                x["_target_column"] = x.get("column")
                x["_selected"] = ""
                self.target_data.append(dict(x))
        self.target_form.set_model(Q2Model())
        self.target_form.model.set_records(self.target_data)
        self.target_form.add_control("target_column", q2app.BULK_TARGET_COLUMNS, control="line", datalen=100)
        self.target_form.add_control("_selected", q2app.BULK_TARGET_SELECTED, control="check", datalen=5)
        self.target_form.grid_double_clicked = self.select
        self.target_form.i_am_child = 1



#    Copyright © 2021 Andrei Puchko
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


from PyQt6.QtWidgets import QLineEdit

from q2gui.pyqt6.q2widget import Q2Widget
from q2gui.q2utils import int_, num
from q2gui import q2app


class q2line(QLineEdit, Q2Widget):
    def __init__(self, meta={}):
        self.formatstring = None
        super().__init__(meta)

        # if self.meta.get("changed"):
        #     self.textChanged.connect(self.meta.get("changed"))

        self.last_text_len = 0
        self.last_cur_pos = 0
        if self.meta.get("pic") == "*":
            self.setEchoMode(self.EchoMode.Password)
        self.TS = " "  # thousands separator
        self.DS = "."  # decimal separator

        if self.meta.get("num"):
            self.declen = int_(self.meta.get("datadec", 0))
            if self.meta.get("pic") == "F":  # financial
                self.formatstring = q2app.FINANCIAL_FORMAT % self.declen
            else:
                self.formatstring = "{:.%sf}" % self.declen
            self.textChanged.connect(self.format_decimal)
            self.cursorPositionChanged.connect(self.track_cursor)
            self.textChanged.emit("")
        else:
            self.setStyleSheet("QLineEdit { qproperty-cursorPosition: 0; }")

    def set_maximum_len(self, length):
        declen = int_(self.meta.get("datadec", 0))
        if self.meta.get("num"):
            add_len = 0 if declen == 0 else 1
            if self.meta.get("pic") == "F":
                add_len += (int(length) - declen) // 3 + 1
            return self.setMaxLength(int(length) + add_len)
        else:
            return super().set_maximum_len(int(length))

    def track_cursor(self, old, new):
        text = self.text()
        cursor_pos = self.cursorPosition()
        if cursor_pos == 0 and len(text) > 0 and text[0] == "-":
            self.setCursorPosition(1)
            new += 1
        self.last_cur_pos = new

    def set_text(self, text):
        if self.formatstring is not None:
            text = self.format_decimal_string(str(text))
        return super().set_text(text)

    def get_text(self):
        if self.meta.get("num"):
            return self.text().replace(self.TS, "")
        else:
            return super().get_text()

    def format_decimal(self):
        text = self.text()
        self.blockSignals(True)
        cursor_pos = self.cursorPosition()
        if cursor_pos == 0 and len(text) > 0 and text[0] == "-":
            cursor_pos = 1
        if text == "." and self.declen > 0:
            cursor_pos = 0

        cursor_pos_right = len(text) - cursor_pos

        negative = text.count("-") == 1 and text.count("+") == 0
        # divide text by cursor pos
        right_text = text[cursor_pos:]
        left_text = text[:-cursor_pos_right] if cursor_pos_right else text

        if self.declen != 0 and self.DS not in text:
            # DS removed
            if cursor_pos_right == self.declen:
                left_text = left_text[:-1] + self.DS
                cursor_pos -= 1
                cursor_pos_right += 1
            # else:
            #     right_text = right_text[1:] + "0"

        if left_text.endswith(","):  # replace comma with point
            left_text = left_text[:-1] + self.DS

        # TS removed
        if left_text and len(text) < self.last_text_len and len(right_text) > 3:
            if (
                right_text[3] in [self.TS, self.DS]
                and left_text[-1] != self.TS
                and cursor_pos - self.last_cur_pos < 0
            ):
                left_text = left_text[:-1]  # remove TS
        # remove non-digit chars and change cursor position
        left_text = "".join([x for x in left_text if x in "0123456789., "])
        text = left_text + right_text

        if self.declen == 0:
            text = text.replace(".", "")
        else:  # for decimal only
            if self.DS not in text:
                text += "."
                cursor_pos_right = self.declen

            if text == "":  # empty text - just 0
                text = "0." + "0" * self.declen
            elif text.count(self.DS) == 0:  # DS deleted
                text = text[:-2] + self.DS + text[-2:]
            elif text.count(self.DS) > 1:  # entered decimal_separator - move cursor postion
                text = left_text[:-1] + right_text
                if "." in right_text:  # added DS left from DS
                    cursor_pos_right = self.declen
                else:
                    cursor_pos_right = self.declen + 1
            else:
                spl = text.split(self.DS)
                if cursor_pos_right <= self.declen:  # cursor in decimal part
                    if len(spl[1]) < self.declen:  # deleted from decimal part
                        text += "0"
                        cursor_pos_right += 1
                    elif len(spl[1]) > self.declen:  # decimal part entered
                        text = text[: -(len(spl[1]) - self.declen)]
                        cursor_pos_right -= 1
                else:  # cursor left of DS
                    if len(spl[0]) == 1 and spl[0][-1] == "0":
                        # 0 only was left from DS-cut 0
                        text = spl[0][:-1] + self.DS + spl[1]
                        cursor_pos_right -= 1
        text = self.format_decimal_string(text)
        cursor_pos = len(text) - cursor_pos_right
        if negative:
            text = f"-{text}"
            cursor_pos += 1
        self.setText(text)
        self.setCursorPosition(cursor_pos)
        self.last_text_len = len(text)
        self.last_cur_pos = cursor_pos
        self.blockSignals(False)

    def format_decimal_string(self, text):
        return self.formatstring.format(num(text.replace(self.TS, ""))).replace(",", self.TS)



#    Copyright © 2021 Andrei Puchko
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


from PyQt6.QtWidgets import (
    QComboBox,
    QWidget,
    QVBoxLayout,
    QCalendarWidget,
    QHBoxLayout,
    QPushButton,
)

from PyQt6.QtGui import QValidator, QFontMetrics
from PyQt6.QtCore import QEvent, pyqtSignal, QDate, Qt

from q2gui.pyqt6.q2widget import Q2Widget
import re
import datetime


class q2date(QComboBox, Q2Widget):
    editingFinished = pyqtSignal()

    def __init__(self, meta):
        super().__init__(meta)
        self.setEditable(True)
        self.lineedit = self.lineEdit()
        if self.meta.get("readonly"):
            self.lineedit.setReadOnly(True)
        self.lineedit.setInputMask("99.99.9999")
        self.set_text(self.meta.get("data"))

        class q2DateValidator(QValidator):
            def validate(self, text, pos):
                if text == "  .  .    ":
                    return (QValidator.State.Acceptable, text, 0)
                else:
                    lt = [x for x in text.replace(" ", "0").split(".")]
                    if int(lt[2]) == 0:  # year
                        lt[2] = QDate.currentDate().toString("yyyy")

                    if int(lt[1]) == 0 and pos != 4:  # month
                        lt[1] = QDate.currentDate().toString("MM")
                    elif int(lt[1]) > 12:
                        lt[1] = "12"

                    if int(lt[0]) > 31:  # day
                        lt[0] = "31"
                    elif int(lt[0]) == 0:
                        lt[0] = "01"
                    elif int(lt[0]) < 4 and len(lt[0]) > 2:
                        lt[0] = f"{int(lt[0])}0"
                    
                    if pos > 9:
                        mdm = QDate(int(lt[2]), int(lt[1]), 1).daysInMonth()
                        if mdm < int(lt[0]):
                            lt[0] = str(mdm)
                    text = ".".join(lt)
                    return (QValidator.State.Acceptable, text, pos)

        self.lineedit.setValidator(q2DateValidator())
        self.lineedit.editingFinished.connect(self.lineeditEditingFinished)
        self.lineedit.cursorPositionChanged.connect(self.lineeditCursorPositionChanged)

    def changeEvent(self, e):
        if e.type() == QEvent.Type.StyleChange:
            self.setMinimumWidth(int(QFontMetrics(self.font()).horizontalAdvance("0") * 16))
        return super().changeEvent(e)

    def lineeditCursorPositionChanged(self, old, new):
        if old in [3, 4] and (new < 3 or new > 4):
            # self.lineedit.setText(self.fixupDay(self.lineedit.text()))
            self.lineedit.setCursorPosition(new)
        elif new == 10:
            self.lineedit.setCursorPosition(9)

    def fixupDay(self, text):
        if text != "..":
            lt = text.split(".")
            mdm = QDate(int(lt[2]), int(lt[1]), 1).daysInMonth()
            if mdm < int(lt[0]):
                lt[0] = str(mdm)
            return ".".join(lt)
        else:
            return text

    def lineeditEditingFinished(self):
        if self.lineedit.text() != "..":
            self.lineedit.setText(self.fixupDay(self.lineedit.text()))
        self.editingFinished.emit()

    def keyPressEvent(self, event):
        if event.key() == Qt.Key.Key_Space:
            self.lineedit.setText("  .  .    ")
        elif event.key() == Qt.Key.Key_End:
            self.lineedit.setCursorPosition(9)
        elif event.key() == Qt.Key.Key_Insert:
            self.showPopup()
            self.lineedit.setCursorPosition(0)
        elif event.key() in [
            Qt.Key.Key_Enter,
            Qt.Key.Key_Return,
            Qt.Key.Key_Down,
            Qt.Key.Key_PageDown,
            Qt.Key.Key_PageUp,
        ]:
            event.ignore()
            # QApplication.sendEvent(self, QKeyEvent(QEvent.Type.KeyPress, Qt.Key.Key_Tab, Qt.NoModifier))
        elif event.key() in [Qt.Key.Key_Up]:
            event.accept()
            self.focusPreviousChild()
        else:
            return super().keyPressEvent(event)

    def showPopup(self):
        if self.meta.get("readonly"):
            return

        class q2CalendarWidget(QWidget):
            def __init__(self, parent=self):
                super().__init__(parent=parent, flags=Qt.WindowType.Popup)
                self.setLayout(QVBoxLayout())
                self.clndr = QCalendarWidget(self)
                self.clndr.setSelectedDate(QDate.fromString(self.parent().lineedit.text(), "dd.MM.yyyy"))
                self.clndr.activated.connect(self.clndrActivated)
                buttonLayout = QHBoxLayout()

                todayButton = QPushButton("Today")
                todayButton.clicked.connect(self.today)
                buttonLayout.addWidget(todayButton)

                clearButton = QPushButton("Clear")
                clearButton.clicked.connect(self.clear)
                buttonLayout.addWidget(clearButton)

                self.layout().addLayout(buttonLayout)
                self.layout().addWidget(self.clndr)

            def show(self):
                super().show()
                self.clndr.setFocus()

            def clear(self):
                self.parent().lineedit.setText("  .  .    ")
                self.close()

            def today(self):
                self.parent().lineedit.setText(QDate.currentDate().toString("dd.MM.yyyy"))
                self.close()

            def clndrActivated(self, date):
                self.parent().set_text(date.toString("yyyy-MM-dd"))
                clndr.close()

        clndr = q2CalendarWidget()
        clndr.setFont(self.font())
        clndr.move(self.mapToGlobal(self.rect().bottomLeft()))
        clndr.show()

    def set_text(self, text):
        if hasattr(self, "lineedit"):
            self.lineedit.setText(extract_date(text))
            # self.lineedit.setText(QDate.fromString(f"{text}", "yyyy-MM-dd").toString("dd.MM.yyyy"))
            self.lineedit.setCursorPosition(0)

    def get_text(self):
        if self.lineedit.text() == "..":
            return "0000-00-00"
        else:
            return QDate.fromString(self.lineedit.text(), "dd.MM.yyyy").toString("yyyy-MM-dd")


def extract_date(text):
    if text == "":
        return ""
    elif text == "0000-00-00":
        return "  .  .    "
    splited_text = re.split(r"[/,/.\s/-]+", text.strip())

    today = datetime.date.today()
    month = None
    year = None
    day = 1
    if len(splited_text) >= 1:  # day
        if len(splited_text[0]) == 4:
            year = splited_text[0]
        elif len(splited_text[0]) < 3:
            day = splited_text[0]

    if len(splited_text) > 1:
        # if month > int(spl[1]):
        #     year = year - 1
        month = int(splited_text[1])
        if month > today.month and len(splited_text) < 3 and year is None:
            year = today.year - 1

    if len(splited_text) > 2:
        if len(splited_text[2]) == 4:
            year = splited_text[2]
        elif len(splited_text[2]) < 3 and len(splited_text[0]) == 4:
            day = splited_text[2]
        else:
            day = splited_text[0]
    if year is None and len(splited_text) == 3:
        if len(splited_text[2]) == 1:
            year = int(splited_text[2]) + today.year // 10 * 10
        else:
            year = int(splited_text[2]) + 2000

    if year is None:
        year = today.year if month != 12 else (today.year - 1)
    if month is None:
        month = (today.month - 1) if today.month > 1 else 12
    date = datetime.date(day=int(day), month=int(month), year=int(year))
    return date.strftime("%d-%m-%Y")
